from enum import Enum


class BarcodePosition(Enum):
  LEFT = "LEFT"
  RIGHT = "RIGHT"


class FilterType(Enum):
  UNDEFINED = "UNDEFINED"
  SHORTPASS = "SP"
  LONGPASS = "LP"
  BANDPASS = "BP"
  OPTICAL_DENSITY = "OD"
  EMPTY = "EMPTY"
  DARK = "DARK"
  AUTOMATIC = "AUTOMATIC"


class FluorescenceCarrier(Enum):
  FILTER_EXCITATION = "FILTER_EX"
  FILTER_EMISSION = "FILTER_EM1"
  FILTER_DUAL_EMISSION = "FILTER_EM2"
  MONOCHROMATOR = "MONO"


class FluorescenceConfiguration(Enum):
  UNKNOWN = "UNKNOWN"
  FILTER_EX_FILTER_EM = "FILTER_EX_FILTER_EM"
  FILTER_EX_MONO_EM = "FILTER_EX_MONO_EM"
  MONO_EX_FILTER_EM = "MONO_EX_FILTER_EM"
  MONO_EX_MONO_EM = "MONO_EX_MONO_EM"


class FluorescenceMeasurementDirection(Enum):
  TOP = "FITOP"
  BOTTOM = "FIBOTTOM"


class InjectionMode(Enum):
  RINSE = "RINSE"
  PRIME = "PRIME"
  DISPENSE = "DISPENSE"
  BACKFLUSH = "BACKFLUSH"
  REFILL = "REFILL"


class InjectorModel(Enum):
  STANDARD = "STANDARD"
  ENHANCED = "ENHANCED"


class InjectorName(Enum):
  A = "A"
  B = "B"
  C = "C"


class InjectorRefillMode(Enum):
  STANDARD = "STANDARD"
  BEFORE_EVERY_INJECTION = "BEFORE_EVERY_INJECTION"


class InjectorState(Enum):
  ACTIVE = "ACTIVE"
  INACTIVE = "INACTIVE"


class LightingName(Enum):
  AUTOFOCUS = "AF"
  BRIGHTFIELD = "BF"
  BLUE = "UV_BLUE"
  GREEN = "BLUE_GREEN"
  RED = "LIME_RED"
  FAR_RED = "RED_FARRED"


class LightingType(Enum):
  AUTOFOCUS = "AF_LED"
  BRIGHTFIELD = "BF_LED"
  FLUORESCENCE = "FI_LED"


class LuminescenceType(Enum):
  NONE = "NONE"
  STANDARD = "STANDARD"
  ENHANCED = "ENHANCED"
  ALPHA = "ALPHA"


class MeasurementMode(Enum):
  ABSORBANCE = "ABS"
  CUVETTE = "CUV"
  LUMINESCENCE = "LUM"
  ALPHA = "ALPHA"
  FLUORESCENCE_TOP = "FITOP"
  FLUORESCENCE_BOTTOM = "FIBOTTOM"
  FLUORESCENCE_POLARIZATION = "FP"
  CELL = "CELL"
  INJECTOR = "INJ"
  WELL_TEMPERATURE = "WELL_TEMP"
  BARCODE = "BARCODE"
  FLUORESCENCE_IMAGING = "FIM"


class MirrorType(Enum):
  HALF_HALF = "50_50"
  DICHROIC = "DICHROIC"
  BOTTOM = "BOTTOM"
  UNUSED = "UNUSED"
  AUTOMATIC = "AUTOMATIC"


class MtpAreaType(Enum):
  SMALL = "SMALL"
  MEDIUM = "MEDIUM"
  LARGE = "LARGE"


class ObjectiveType(Enum):
  UNDEFINED = "XX"
  TWO_TIMES = "2X"
  FOUR_TIMES = "4X"
  TEN_TIMES = "10X"


class PlatePosition(Enum):
  UNDEFINED = "UNDEFINED"
  OUT_LEFT = "OUT_LEFT"
  OUT_RIGHT = "OUT_RIGHT"
  PLATE_IN = "PLATE_IN"
  PICK_N_PLACE = "PICK_N_PLACE"
  LID_LIFTER = "LIDLIFTER"
  CHECK = "CHECK"
  HEATING = "HEATING"
  INCUBATION = "INCUBATION"
  COOLING = "COOLING"
  BARCODE_LEFT = "BARCODE_LEFT"
  BARCODE_RIGHT = "BARCODE_RIGHT"


class ShakingMode(Enum):
  LINEAR = "LINEAR"
  ORBITAL = "ORBITAL"
  DOUBLE = "DOUBLE"


class ShakingName(Enum):
  SLOW = "SLOW"
  MEDIUM = "MEDIUM"
  FAST = "FAST"


class ColumnType(Enum):
  INPUT = "INPUT"
  OUTPUT = "OUTPUT"


class HumidityCassetteModel(Enum):
  STANDARD = "STANDARD"
  CYTO = "CYTO"


class InstrumentMessageType(Enum):
  ALL = "ALL"
  TEMPERATURE = "TEMPERATURE"


class MovementSpeed(Enum):
  NORMAL = "NORMAL"
  SMALLSTEP = "SMALLSTEP"
  AUTOMATIC = "AUTOMATIC"
  SMOOTH = "SMOOTH"
  SLOW = "SLOW"
  FLASH_INJECTION = "FLASH_INJECTION"


class StackerColumn(Enum):
  PRESENT = "PRESENT"
  NOT_PRESENT = "NOT_PRESENT"
  UNKNOWN = "UNKNOWN"


class StackerPlate(Enum):
  PRESENT = "PRESENT"
  NOT_PRESENT = "NOT_PRESENT"
  UNKNOWN = "UNKNOWN"


class ControlCommandType(Enum):
  OFF = "OFF"
  ON = "ON"
  TARGET_TEMP = "TARGET_TEMP"


class GasBuzzerState(Enum):
  OFF = "OFF"
  ON = "ON"


class GasMode(Enum):
  OFF = "OFF"
  LOG = "LOG"
  CTRL = "CTRL"


class GasOption(Enum):
  O2 = "O2"
  CO2 = "CO2"


class GasPowerState(Enum):
  OFF = "OFF"
  ON = "ON"


class GasSensorState(Enum):
  OFF = "OFF"
  INIT = "INIT"
  READY = "READY"


class GasErrorMessage(Enum):
  TIMEOUT = 1240
  SENSOR_HEATS_UP = 1241
  SENSOR_SIGNAL_CORRUPT = 1242
  GAS_CONFIG_ERROR = 1243


class GasMessage(Enum):
  CO2_CONCENTRATION = 250
  O2_CONCENTRATION = 251
  SENSOR_READY = 252


class SymbioErrors(Enum):
  TEMPERATURE_SENSOR_READING = 1104
  INJECTOR_CARRIER_INSERTED = 1213
  GAS_TIMEOUT = 1240
  GAS_SIGNAL_CORRUPT = 1242


class SymbioMessages(Enum):
  TEMPERATURE_CHAMBER = 100
  TEMPERATURE_AMBIENT = 101
  EXCITATION_FILTERSLIDE_INSERTED = 150
  EMISSION_FILTERSLIDE_INSERTED = 151
  DUAL_EMISSION_FILTERSLIDE_INSERTED = 152
  POWER_ON = 200
  POWER_OFF = 201
  START_STOP = 202
  FILTERSLIDES_OUT = 203
  PLATE_OUT_IN = 204
  GAS_CO2_CONCENTRATION = 250
  GAS_O2_CONCENTRATION = 251
  GAS_SENSOR_READY = 252
  INJECTOR_A_PRIME = 300
  INJECTOR_A_RINSE = 301
  INJECTOR_B_PRIME = 302
  INJECTOR_B_RINSE = 303
  INJECTOR_C_PRIME = 304
  INJECTOR_C_RINSE = 305
  INJECTOR_ACTION_FINISHED = 306
  STACKER_INPUT_COLUMN_PRESENT = 401
  STACKER_INPUT_COLUMN_NOT_PRESENT = 402
  STACKER_OUTPUT_COLUMN_PRESENT = 403
  STACKER_OUTPUT_COLUMN_NOT_PRESENT = 404


class Mode(Enum):
  BOOT = "B"
  SERVICE = "S"
  OPERATION = "O"


class ModuleName(Enum):
  BOOT = "BOOT"
  MAIN = "MAIN"
  LUM = "LUM"
  MEX = "MEX"
  MEM = "MEM"
  FIM = "FIM"


class ModuleType(Enum):
  BOOTLOADER = "BOOT"
  MAIN = "MAIN"
  LUMINESCENCE = "LUM"
  MONO_EXCITATION = "MEX"
  MONO_EMISSION = "MEM"
  FLUORESCENCE_IMAGING = "FIM"
  ABSORBANCE = "ABS"
  FLUORESCENCE = "FLUOR"
  PLATE_TRANSPORT = "MTP"
  STACKER = "STACKER"
  INJECTOR = "INJ"
  POWER_DISTRIBUTION_BOARD = "PODI"
  COOLING = "COOLING"
  BARCODE = "BARCODE"
  CELL = "CELL"
  GAS_CONTROL = "GCM"
  LIFTER = "LIFT"
  PICK_AND_PLACE = "PAP"
  TWO_STEP_MOTORS = "2SM"
  USB_CAMERA = "USBCAM"
  USB_CAMERA2 = "USBCAM2"


class CameraMode(Enum):
  PREPARE = "PREPARE"
  TRIGGER = "TRIGGER"
  VIDEO = "VIDEO"


class GasMessageType(Enum):
  CONCENTRATION = "CONCENTRATION"
  SENSOR_READY = "SENSOR_READY"
  GAS_ERROR = "GAS_ERROR"


class HardwareButtons(Enum):
  START_STOP = "START_STOP"
  FILTER_OUT = "FILTER_OUT"
  PLATE = "PLATE"
  INJECTOR = "INJECTOR"
  POWER = "POWER"
  ALL = "ALL"


class InjectorCarrierState(Enum):
  OPERATING_POSITION = "OPERATINGPOSITION"
  PRIME_POSITION = "PRIMEPOSITION"


class LedColor(Enum):
  RED = "RED"
  GREEN = "GREEN"
  BLUE = "BLUE"
  YELLOW = "YELLOW"
  MAGENTA = "MAGENTA"
  CYAN = "CYAN"
  WHITE = "WHITE"
  BLACK = "BLACK"


class LedState(Enum):
  OFF = "OFF"
  STANDBY = "STANDBY"
  IDLE = "IDLE"
  IDLE_ACQUIRED = "IDLE_ACQUIRED"
  RUN = "RUN"
  ERROR = "ERROR"
  USER_INTERACTION = "USER_INTERACTION"
  PAUSE = "PAUSE"
  SHUT_DOWN = "SHUT_DOWN"
  ACTION_IMPOSSIBLE = "ACTION_IMPOSSIBLE"


class MotorDirection(Enum):
  LEFT = "LEFT"
  RIGHT = "RIGHT"


class MoveableCarrier(Enum):
  EXCITATION_FILTER = "EXCITATION_FILTER"
  EMISSION_FILTER = "EMISSION_FILTER"
  DUAL_PMT_EMISSION_FILTER = "DUAL_PMT_EMISSION_FILTER"
  MIRROR = "MIRROR"
  DUAL_PMT_MIRROR = "DUAL_PMT_MIRROR"
  ALL = "ALL"


class MoveableCarrierPosition(Enum):
  IN = "IN"
  OUT = "OUT"


class MtpMotor(Enum):
  X = "X"
  Y = "Y"
  Z = "Z"


class PolarisationEmissionMode(Enum):
  AUTOMATIC = "AUTOMATIC"
  MANUAL = "MANUAL"


class PolarisationMode(Enum):
  PARALLEL = "PARALLEL"
  PERPENDICULAR = "PERPENDICULAR"


class Retractable(Enum):
  FILTER_SLIDE = "FILTER_SLIDE"


class State(Enum):
  ON = "ON"
  OFF = "OFF"


class TargetMode(Enum):
  AMBIENT = "AMBIENT"
  FIX = "FIX"


class SoftwareCounter(Enum):
  PLATE_0001_WELL = "PLATE_0001_WELL"
  ABSORBANCE = "ABSORBANCE"
  ABSORBANCE_SCAN = "ABSORBANCE_SCAN"
  FLUORESCENCE_INTENSITY = "FLUORESCENCE_INTENSITY"
  FLUORESCENCE_INTENSITY_SCAN = "FLUORESCENCE_INTENSITY_SCAN"
  FLUORESCENCE_POLARIZATION = "FLUORESCENCE_POLARIZATION"
  LUMINESCENCE = "LUMINESCENCE"
  LUMINESCENCE_SCAN = "LUMINESCENCE_SCAN"
  ALPHA = "ALPHA"
  CELL = "CELL"
  FLASH_DROPOUT = "FLASH_DROPOUT"
  ON_BOARD_START = "ON_BOARD_START"
  ON_BOARD_CONTINUE = "ON_BOARD_CONTINUE"
  ON_BOARD_STOP = "ON_BOARD_STOP"


class FirmwareCounter(Enum):
  LID_LIFTED = "LID_LIFTED"
  X_MOVEMENT = "X_MOVEMENT"
  Y_MOVEMENT = "Y_MOVEMENT"
  Z_MOVEMENT = "Z_MOVEMENT"
  SHAKING = "SHAKING"
  HEATING_STANDARD = "HEATING_STANDARD"
  HEATING_ENHANCED = "HEATING_ENHANCED"
  INSTRUMENT_ON_TIME = "INSTRUMENT_ON_TIME"
  FLASHES = "FLASHES"
  LASER_ON_TIME = "LASER_ON_TIME"
  VALVE_SWITCHES_O2 = "VALVE_SWITCHES_O2"
  VALVE_SWITCHES_CO2 = "VALVE_SWITCHES_CO2"
  PUMPED_VOLUME_INJECTOR_A = "PUMPED_VOLUME_INJECTOR_A"
  PUMPED_VOLUME_INJECTOR_B = "PUMPED_VOLUME_INJECTOR_B"
  PUMPED_VOLUME_INJECTOR_C = "PUMPED_VOLUME_INJECTOR_C"
  ON_TIME = "ON_TIME"
  VALVE_SWITCHES = "VALVE_SWITCHES"
  READING = "READING"


class AreaOfInterestProperty(Enum):
  X = "X"
  Y = "Y"
  WIDTH = "WIDTH"
  HEIGHT = "HEIGHT"


class CameraExecutionDetails(Enum):
  SUCCESSFUL = "SUCCESSFUL"
  FRAME_DROP = "FRAME_DROP"


class InstrumentMode(Enum):
  OPERATION = "OPERATION"
  SERVICE = "SERVICE"


class Unit(Enum):
  NONE = "NONE"
  S = "S"
  NS = "NS"
  US_MICRO = "µS"
  US = "US"
  UM = "UM"
  M = "M"
  PERCENT = "PERCENT"
  STEP = "STEP"
  MS = "MS"
  FPS = "FPS"
  MHZ = "MHZ"
  HZ10 = "HZ10"
  ANG = "ANG"
  HZ = "HZ"
  C100 = "C100"
  UL_MICRO = "µL"
  UL = "UL"
  UL_PER_S = "UL_PER_S"
  DHZ = "DHZ"
  PPM = "PPM"
  H = "H"
  ML = "ML"
  KHZ = "KHZ"


# Limits Enums - Using FirmwareNames as values where available
class AbsorbanceLimit(Enum):
  REFERENCE_DATA_RANGE_LOW = "REF_DATARANGE_LOW"
  SIGNAL_DATA_RANGE_LOW = "SIG_DATARANGE_LOW"
  REFERENCE_DATA_RANGE_HIGH = "REF_DATARANGE_HIGH"
  SIGNAL_DATA_RANGE_HIGH = "SIG_DATARANGE_HIGH"
  OD_MAX = "OD_MAX"
  FLASH_DROPOUT = "FLASH_DROPOUT"


class CameraLimit(Enum):
  ROI_START_X = "ROI_START_X"
  ROI_START_Y = "ROI_START_Y"
  LASER_OFFSET = "LASER_OFFSET"
  INT_TIME_THIN = "INT_TIME_THIN"
  INT_TIME_THICK = "INT_TIME_THICK"


class FluorescenceLimit(Enum):
  REFERENCE_FILTER_RANGE_LOW = "REF_FIL_RANGE_LOW"
  REFERENCE_FILTER_RANGE_HIGH = "REF_FIL_RANGE_HIGH"
  REFERENCE_MONO_RANGE_LOW = "REF_MONO_RANGE_LOW"
  REFERENCE_MONO_RANGE_HIGH = "REF_MONO_RANGE_HIGH"
  SIGNAL_RANGE_LOW = "SIGNAL_RANGE_LOW"
  SIGNAL_RANGE_HIGH = "SIGNAL_RANGE_HIGH"
  FLASH_DROPOUT = "FLASH_DROPOUT"
  EX_WL_FILTER_TOP_MIN = "EX_WL_FIL_TOP_MIN"
  EX_WL_FILTER_TOP_MAX = "EX_WL_FIL_TOP_MAX"
  EM_WL_FILTER_TOP_MIN = "EM_WL_FIL_TOP_MIN"
  EM_WL_FILTER_TOP_MAX = "EM_WL_FIL_TOP_MAX"
  EX_WL_FILTER_BOTTOM_MIN = "EX_WL_FIL_BOTTOM_MIN"
  EX_WL_FILTER_BOTTOM_MAX = "EX_WL_FIL_BOTTOM_MAX"
  EM_WL_FILTER_BOTTOM_MIN = "EM_WL_FIL_BOTTOM_MIN"
  EM_WL_FILTER_BOTTOM_MAX = "EM_WL_FIL_BOTTOM_MAX"
  EX_WL_FP_MIN = "EX_WL_FP_MIN"
  EX_WL_FP_MAX = "EX_WL_FP_MAX"
  EM_WL_FP_MIN = "EM_WL_FP_MIN"
  EM_WL_FP_MAX = "EM_WL_FP_MAX"
  EX_WL_MONO_MIN = "EX_WL_MONO_MIN"
  EX_WL_MONO_MAX = "EX_WL_MONO_MAX"
  EM_WL_MONO_MIN = "EM_WL_MONO_MIN"
  EM_WL_MONO_MAX = "EM_WL_MONO_MAX"
  MAX_CHAR_FILTER_DESCRIPTION = "MAX_CHAR_FILTERDESCR"


class LuminescenceLimit(Enum):
  DARK_TIME_MIN = "DARKTIMEMIN"
  LID_CHECK_MAX = "LIDCHECKMAX"
  DARK_MAX = "DARKMAX"
  DARK_TIME_MAX = "DARKTIMEMAX"
  ALPHA_TEMP_SLOPE = "TEMP_SLOPE"
  ALPHA_TEMP_INTERCEPT = "TEMP_INTERCEPT"


class PlateTransportLimit(Enum):
  MIN_PLATE_HEIGHT = "MIN_PLATEHEIGHT"
  MAX_PLATE_HEIGHT = "MAX_PLATEHEIGHT"
  PLATE_REFERENCE_WIDTH = "PLATE_REF_WIDTH"
  MAX_PLATE_FORMAT = "MAX_PLATEFORMAT"
  DELTA_WELL_TEMP_LUM_X = "DELTAX_WELLTEMP_LUM"
  DELTA_WELL_TEMP_LUM_Y = "DELTAY_WELLTEMP_LUM"


class ConfigAxis(Enum):
  X = "X"
  Y = "Y"
  Z = "Z"


class TriggerMode(Enum):
  SOFTWARE = "SOFTWARE"
  EXTERNAL = "EXTERNAL"


class FlippingMode(Enum):
  STANDARD = "STANDARD"
  HORIZONTAL = "HORIZONTAL"
  VERTICAL = "VERTICAL"
  BOTH = "BOTH"


class BrightnessState(Enum):
  ENABLED = "ENABLED"
  DISABLED = "DISABLED"


class TemperatureDevice(Enum):
  PLATE = "PLATE"


class TemperatureState(Enum):
  ON = "ON"
  OFF = "OFF"


class ChillerState(Enum):
  ON = "ON"
  OFF = "OFF"


class MirrorCarrier(Enum):
  MIRROR1 = "MIRROR1"


class LaserPowerState(Enum):
  ON = "ON"
  OFF = "OFF"


class LightingState(Enum):
  ON = "ON"
  OFF = "OFF"


class LidLiftState(Enum):
  ON = "ON"
  OFF = "OFF"


class RetractionState(Enum):
  ENABLED = "ENABLED"
  DISABLED = "DISABLED"


class PlateColor(Enum):
  BLACK = "BLACK"
  WHITE = "WHITE"
  TRANSPARENT = "TRANSPARENT"
  NO = "NO"


class ScanDirection(Enum):
  UP = "UP"
  DOWN = "DOWN"
  ALTERNATE_UP = "ALTERNATE_UP"
  ALTERNATE_DOWN = "ALTERNATE_DOWN"


class ScanDarkState(Enum):
  TRUE = "TRUE"
  FALSE = "FALSE"


class SimulationState(Enum):
  ON = "ON"
  OFF = "OFF"



import asyncio
import warnings
from abc import ABC
from typing import Dict, List, Literal, Optional, Union

from pylabrobot.arms.backend import (
  AccessPattern,
  HorizontalAccess,
  SCARABackend,
  VerticalAccess,
)
from pylabrobot.arms.precise_flex.coords import ElbowOrientation, PreciseFlexCartesianCoords
from pylabrobot.arms.precise_flex.error_codes import ERROR_CODES
from pylabrobot.arms.precise_flex.joints import PFAxis
from pylabrobot.io.socket import Socket
from pylabrobot.resources import Coordinate, Rotation


class PreciseFlexError(Exception):
  def __init__(self, replycode: int, message: str):
    self.replycode = replycode
    self.message = message

    # Map error codes to text and descriptions
    error_info = ERROR_CODES
    if replycode in error_info:
      text = error_info[replycode]["text"]
      description = error_info[replycode]["description"]
      super().__init__(f"PreciseFlexError {replycode}: {text}. {description} - {message}")
    else:
      super().__init__(f"PreciseFlexError {replycode}: {message}")


class PreciseFlexBackend(SCARABackend, ABC):
  """Backend for the PreciseFlex robotic arm  - Default to using Cartesian coordinates, some methods in Brook's TCS don't work with Joint coordinates.

  Documentation and error codes available at https://www2.brooksautomation.com/#Root/Welcome.htm
  """

  def __init__(
    self,
    host: str,
    port: int = 10100,
    is_dual_gripper: bool = False,
    has_rail: bool = False,
    timeout=20,
  ) -> None:
    super().__init__()
    self.io = Socket(human_readable_device_name="Precise Flex Arm", host=host, port=port)
    self.profile_index: int = 1
    self.location_index: int = 1
    self.horizontal_compliance: bool = False
    self.horizontal_compliance_torque: int = 0
    self.timeout = timeout
    self._has_rail = has_rail
    self._is_dual_gripper = is_dual_gripper
    if is_dual_gripper:
      warnings.warn(
        "Dual gripper support is experimental and may not work as expected.", UserWarning
      )

  def _convert_to_cartesian_space(
    self, position: tuple[float, float, float, float, float, float, Optional[ElbowOrientation]]
  ) -> PreciseFlexCartesianCoords:
    """Convert a tuple of cartesian coordinates to a CartesianCoords object."""
    if len(position) != 7:
      raise ValueError(
        "Position must be a tuple of 7 values (x, y, z, yaw, pitch, roll, orientation)."
      )
    orientation = ElbowOrientation(position[6])
    return PreciseFlexCartesianCoords(
      location=Coordinate(position[0], position[1], position[2]),
      rotation=Rotation(position[5], position[4], position[3]),
      orientation=orientation,
    )

  def _convert_to_cartesian_array(
    self, position: PreciseFlexCartesianCoords
  ) -> tuple[float, float, float, float, float, float, int]:
    """Convert a CartesianCoords object to a list of cartesian coordinates."""
    orientation_int = self._convert_orientation_enum_to_int(position.orientation)
    arr = (
      position.location.x,
      position.location.y,
      position.location.z,
      position.rotation.yaw,
      position.rotation.pitch,
      position.rotation.roll,
      orientation_int,
    )
    return arr

  async def setup(self, skip_home: bool = False):
    """Initialize the PreciseFlex backend."""
    await self.io.setup()
    await self.set_response_mode("pc")
    await self.power_on_robot()
    await self.attach(1)
    if not skip_home:
      await self.home()

  async def stop(self):
    """Stop the PreciseFlex backend."""
    await self.detach()
    await self.power_off_robot()
    await self.exit()
    await self.io.stop()

  async def set_speed(self, speed_percent: float):
    """Set the speed percentage of the arm's movement (0-100)."""
    await self.set_profile_speed(self.profile_index, speed_percent)

  async def get_speed(self) -> float:
    """Get the current speed percentage of the arm's movement."""
    return await self.get_profile_speed(self.profile_index)

  async def open_gripper(self, gripper_width: float):
    """Open the gripper to the specified width. If no width is specified, opens to the default open position."""
    await self._set_grip_open_pos(gripper_width)
    await self.send_command("gripper 1")

  async def close_gripper(self, gripper_width: float):
    """Close the gripper to the specified width. If no width is specified, closes to the default close position."""
    await self._set_grip_close_pos(gripper_width)
    await self.send_command("gripper 2")

  async def halt(self):
    """Stops the current robot immediately but leaves power on."""
    await self.send_command("halt")

  async def home(self) -> None:
    """Home the robot associated with this thread.

    Note:
      Requires power to be enabled.
      Requires robot to be attached.
      Waits until the homing is complete.
    """
    await self.send_command("home")

  async def move_to_safe(self) -> None:
    """Moves the robot to Safe Position.

    Does not include checks for collision with 3rd party obstacles inside the work volume of the robot.
    """
    await self.send_command("movetosafe")

  def _convert_orientation_int_to_enum(self, orientation_int: int) -> Optional[ElbowOrientation]:
    if orientation_int == 1:
      return ElbowOrientation.RIGHT
    if orientation_int == 2:
      return ElbowOrientation.LEFT
    return None

  def _convert_orientation_enum_to_int(self, orientation: Optional[ElbowOrientation]) -> int:
    if orientation == ElbowOrientation.LEFT:
      return 2
    if orientation == ElbowOrientation.RIGHT:
      return 1
    return 0

  async def home_all(self) -> None:
    """Home all robots.

    Note:
      Requires power to be enabled.
      Requires that robots not be attached.
    """
    await self.send_command("homeAll")

  async def attach(self, attach_state: Optional[int] = None) -> int:
    """Attach or release the robot, or get attachment state.

    Args:
      attach_state: If omitted, returns the attachment state.  0 = release the robot; 1 = attach the robot.

    Returns:
      If attach_state is omitted, returns 0 if robot is not attached, -1 if attached.  Otherwise returns 0 on success.

    Note:
      The robot must be attached to allow motion commands.
    """
    if attach_state is None:
      response = await self.send_command("attach")
      return int(response)
    await self.send_command(f"attach {attach_state}")
    return 0

  async def detach(self):
    """Detach the robot."""
    await self.attach(0)

  async def power_on_robot(self):
    """Power on the robot."""
    await self.set_power(True, self.timeout)

  async def power_off_robot(self):
    """Power off the robot."""
    await self.set_power(False)

  async def approach(
    self,
    position: Union[PreciseFlexCartesianCoords, Dict[int, float]],
    access: Optional[AccessPattern] = None,
  ):
    """Move the arm to an approach position (offset from target).

    Args:
      position: Target position (CartesianCoords or Dict[int, float])
      access: Access pattern defining how to approach the target. Defaults to VerticalAccess() if not specified.

    Example:
      # Simple vertical approach (default)
      await backend.approach(position)

      # Horizontal hotel-style approach
      await backend.approach(
        position,
        HorizontalAccess(
          approach_distance_mm=50,
          clearance_mm=50,
          lift_height_mm=100
        )
      )
    """
    if access is None:
      access = VerticalAccess()

    if isinstance(position, dict):
      await self._approach_j(position, access)
    elif isinstance(position, PreciseFlexCartesianCoords):
      await self._approach_c(position, access)
    else:
      raise TypeError("Position must be of type Dict[int, float] or CartesianCoords.")

  async def pick_up_resource(
    self,
    position: Union[PreciseFlexCartesianCoords, Dict[int, float]],
    plate_width: float,
    access: Optional[AccessPattern] = None,
    finger_speed_percent: float = 50.0,
    grasp_force: float = 10.0,
  ):
    """Pick a plate from the specified position.

    Args:
      position: Target position for pickup (CartesianCoords only, joint coords not supported)
      plate_width: Gripper width in millimeters used when gripping the plate.
      access: How to access the location (VerticalAccess or HorizontalAccess).  Defaults to VerticalAccess() if not specified.
      finger_speed_percent: Speed percentage for the gripper fingers (1-100)
      grasp_force: Grasp force in Newtons

    Raises:
      ValueError: If position is not CartesianCoords

    Example:
      # Simple vertical pick (default)
      await backend.pick_plate(position)

      # Vertical pick with custom clearance
      await backend.pick_plate(position, VerticalAccess(clearance_mm=150))

      # Horizontal hotel-style pick
      await backend.pick_plate(
        position,
        HorizontalAccess(
          approach_distance_mm=50,
          clearance_mm=50,
          lift_height_mm=100
        )
      )
    """
    if access is None:
      access = VerticalAccess()

    await self.set_grasp_data(
      plate_width=plate_width,
      finger_speed_percent=finger_speed_percent,
      grasp_force=grasp_force,
    )

    if isinstance(position, PreciseFlexCartesianCoords):
      await self._pick_plate_c(cartesian_position=position, access=access)
    elif isinstance(position, dict):
      await self._pick_plate_j(position, access)
    else:
      raise TypeError("Position must be of type Dict[int, float] or CartesianCoords.")

  async def drop_resource(
    self,
    position: Union[PreciseFlexCartesianCoords, Dict[int, float]],
    access: Optional[AccessPattern] = None,
  ):
    """Place a plate at the specified position.

    Args:
      position: Target position for placement (CartesianCoords only, joint coords not supported)
      access: How to access the location (VerticalAccess or HorizontalAccess).  Defaults to VerticalAccess() if not specified.

    Raises:
      ValueError: If position is not CartesianCoords

    Example:
      # Simple vertical place (default)
      await backend.place_plate(position)

      # Vertical place with custom clearance
      await backend.place_plate(position, VerticalAccess(clearance_mm=150))

      # Horizontal hotel-style place
      await backend.place_plate(
        position,
        HorizontalAccess(
          approach_distance_mm=50,
          clearance_mm=50,
          lift_height_mm=100
        )
      )
    """
    if access is None:
      access = VerticalAccess()

    if not isinstance(position, PreciseFlexCartesianCoords):
      raise TypeError("place_plate only supports CartesianCoords for PreciseFlex.")
    await self._place_plate_c(cartesian_position=position, access=access)

  async def move_to(self, position: Union[PreciseFlexCartesianCoords, Dict[int, float]]):
    """Move the arm to a specified position in 3D space.

    Args:
      position: Either CartesianCoords or a dict mapping PFAxis to float values.
        When using a dict, any unspecified axes will be filled in from the current position.
    """
    print(position, isinstance(position, dict))
    if isinstance(position, dict):
      current = await self.get_joint_position()
      joint_coords = {**current, **position}
      await self.move_j(profile_index=self.profile_index, joint_coords=joint_coords)
    elif isinstance(position, PreciseFlexCartesianCoords):
      await self.move_c(profile_index=self.profile_index, cartesian_coords=position)
    else:
      raise TypeError("Position must be of type Dict[int, float] or CartesianCoords.")

  async def get_joint_position(self) -> Dict[int, float]:
    """Get the current joint position of the arm."""

    await self.wait_for_eom()

    num_tries = 2
    for _ in range(num_tries):
      data = await self.send_command("wherej")
      parts = data.split()
      if len(parts) > 0:
        break
    else:
      raise PreciseFlexError(-1, "Unexpected response format from wherej command.")

    return self._parse_angles_response(parts)

  async def get_cartesian_position(self) -> PreciseFlexCartesianCoords:
    """Get the current position of the arm in 3D space."""

    await self.wait_for_eom()

    num_tries = 2
    for _ in range(num_tries):
      data = await self.send_command("wherec")
      parts = data.split()
      if len(parts) == 7:
        break
    else:
      raise PreciseFlexError(-1, "Unexpected response format from wherec command.")

    x, y, z, yaw, pitch, roll = self._parse_xyz_response(parts[0:6])
    config = int(parts[6])

    # return (x, y, z, yaw, pitch, roll, config)
    enum_thing = self._convert_orientation_int_to_enum(config)

    return self._convert_to_cartesian_space(position=(x, y, z, yaw, pitch, roll, enum_thing))

  async def send_command(self, command: str) -> str:
    await self.io.write(command.encode("utf-8") + b"\n")
    reply = await self.io.readline()

    return self._parse_reply_ensure_successful(reply)

  def _parse_reply_ensure_successful(self, reply: bytes) -> str:
    """Parse reply from Precise Flex.

    Expected format: b'replycode data message\r\n'
    - replycode is an integer at the beginning
    - data is rest of the line (excluding CRLF)
    """
    text = reply.decode().strip()  # removes \r\n
    if not text:
      raise PreciseFlexError(-1, "Empty reply from device.")

    parts = text.split(" ", 1)
    if len(parts) == 1:
      replycode = int(parts[0])
      data = ""
    else:
      replycode, data = int(parts[0]), parts[1]

    if replycode != 0:
      # if error is reported, the data part generally contains the error message
      raise PreciseFlexError(replycode, data)

    return data

  async def _approach_j(self, joint_position: Dict[int, float], access: AccessPattern):
    """Move the arm to a position above the specified coordinates.

    The approach behavior depends on the access pattern:
    - VerticalAccess: Approaches from above using approach_height_mm
    - HorizontalAccess: Approaches from the side using approach_distance_mm
    """
    await self.set_joint_angles(self.location_index, joint_position)
    await self._set_grip_detail(access)
    await self.move_to_stored_location_appro(self.location_index, self.profile_index)

  async def _pick_plate_j(self, joint_position: Dict[int, float], access: AccessPattern):
    """Pick a plate from the specified position using joint coordinates."""
    await self.set_joint_angles(self.location_index, joint_position)
    await self._set_grip_detail(access)
    await self.pick_plate_from_stored_position(
      self.location_index, self.horizontal_compliance, self.horizontal_compliance_torque
    )

  async def _place_plate_j(self, joint_position: Dict[int, float], access: AccessPattern):
    """Place a plate at the specified position using joint coordinates."""
    await self.set_joint_angles(self.location_index, joint_position)
    await self._set_grip_detail(access)
    await self.place_plate_to_stored_position(
      self.location_index, self.horizontal_compliance, self.horizontal_compliance_torque
    )

  async def _approach_c(
    self,
    cartesian_position: PreciseFlexCartesianCoords,
    access: AccessPattern,
  ):
    """Move the arm to a position above the specified coordinates.

    The approach behavior depends on the access pattern:
    - VerticalAccess: Approaches from above using approach_height_mm
    - HorizontalAccess: Approaches from the side using approach_distance_mm
    """
    await self.set_location_xyz(self.location_index, cartesian_position)
    await self._set_grip_detail(access)
    orientation_int = self._convert_orientation_enum_to_int(cartesian_position.orientation)
    await self.set_location_config(self.location_index, orientation_int)
    await self.move_to_stored_location_appro(self.location_index, self.profile_index)

  async def _pick_plate_c(
    self,
    cartesian_position: PreciseFlexCartesianCoords,
    access: AccessPattern,
  ):
    """Pick a plate from the specified position using Cartesian coordinates."""
    await self.set_location_xyz(self.location_index, cartesian_position)
    await self._set_grip_detail(access)
    orientation_int = self._convert_orientation_enum_to_int(cartesian_position.orientation)
    orientation_int |= 0x1000  # GPL_Single: restrict wrist to ±180°
    await self.set_location_config(self.location_index, orientation_int)
    await self.pick_plate_from_stored_position(
      self.location_index, self.horizontal_compliance, self.horizontal_compliance_torque
    )

  async def _place_plate_c(
    self,
    cartesian_position: PreciseFlexCartesianCoords,
    access: AccessPattern,
  ):
    """Place a plate at the specified position using Cartesian coordinates."""
    await self.set_location_xyz(self.location_index, cartesian_position)
    await self._set_grip_detail(access)
    orientation_int = self._convert_orientation_enum_to_int(cartesian_position.orientation)
    orientation_int |= 0x1000  # GPL_Single: restrict wrist to ±180°
    await self.set_location_config(self.location_index, orientation_int)
    await self.place_plate_to_stored_position(
      self.location_index, self.horizontal_compliance, self.horizontal_compliance_torque
    )

  async def _set_grip_detail(self, access: AccessPattern):
    """Configure station type for pick/place operations based on access pattern.

    Calls TCS set_station_type command to configure how the robot interprets
    clearance values and performs approach/retract motions.

    Args:
      access: Access pattern (VerticalAccess or HorizontalAccess) defining how to approach and retract from the location.
    """
    if isinstance(access, VerticalAccess):
      # Vertical access: access_type=1, z_clearance is vertical distance
      await self.set_station_type(
        station_id=self.location_index,
        access_type=1,
        location_type=0,
        z_clearance=access.clearance_mm,
        z_above=0,
        z_grasp_offset=access.gripper_offset_mm,
      )
    elif isinstance(access, HorizontalAccess):
      # Horizontal access: access_type=0, z_clearance is horizontal distance
      await self.set_station_type(
        station_id=self.location_index,
        access_type=0,
        location_type=0,
        z_clearance=access.clearance_mm,
        z_above=access.lift_height_mm,
        z_grasp_offset=access.gripper_offset_mm,
      )
    else:
      raise TypeError("Access pattern must be VerticalAccess or HorizontalAccess.")

  # region GENERAL COMMANDS

  async def get_base(self) -> tuple[float, float, float, float]:
    """Get the robot base offset.

    Returns:
      A tuple containing (x_offset, y_offset, z_offset, z_rotation)
    """
    data = await self.send_command("base")
    parts = data.split()
    if len(parts) != 4:
      raise PreciseFlexError(-1, "Unexpected response format from base command.")

    x_offset = float(parts[0])
    y_offset = float(parts[1])
    z_offset = float(parts[2])
    z_rotation = float(parts[3])

    return (x_offset, y_offset, z_offset, z_rotation)

  async def set_base(
    self, x_offset: float, y_offset: float, z_offset: float, z_rotation: float
  ) -> None:
    """Set the robot base offset.

    Args:
      x_offset: Base X offset
      y_offset: Base Y offset
      z_offset: Base Z offset
      z_rotation: Base Z rotation

    Note:
      The robot must be attached to set the base.
      Setting the base pauses any robot motion in progress.
    """
    await self.send_command(f"base {x_offset} {y_offset} {z_offset} {z_rotation}")

  async def exit(self) -> None:
    """Close the communications link immediately.

    Note:
      Does not affect any robots that may be active.
    """
    await self.io.write(b"exit\n")

  async def get_power_state(self) -> int:
    """Get the current robot power state.

    Returns:
      Current power state (0 = disabled, 1 = enabled)
    """
    response = await self.send_command("hp")
    return int(response)

  async def set_power(self, enable: bool, timeout: int = 0) -> None:
    """Enable or disable robot high power.

    Args:
      enable: True to enable power, False to disable
      timeout: Wait timeout for power to come on.
        0 or omitted = do not wait for power to come on
        > 0 = wait this many seconds for power to come on
        -1 = wait indefinitely for power to come on

    Raises:
      PreciseFlexError: If power does not come on within the specified timeout.
    """
    power_state = 1 if enable else 0
    if timeout == 0:
      await self.send_command(f"hp {power_state}")
    else:
      await self.send_command(f"hp {power_state} {timeout}")

  ResponseMode = Literal["pc", "verbose"]

  async def get_mode(self) -> ResponseMode:
    """Get the current response mode.

    Returns:
      Current mode (0 = PC mode, 1 = verbose mode)
    """
    response = await self.send_command("mode")
    mapping: Dict[int, PreciseFlexBackend.ResponseMode] = {
      0: "pc",
      1: "verbose",
    }
    return mapping[int(response)]

  async def set_response_mode(self, mode: ResponseMode) -> None:
    """Set the response mode.

    Args:
      mode: Response mode to set.
      0 = Select PC mode
      1 = Select verbose mode

    Note:
      When using serial communications, the mode change does not take effect
      until one additional command has been processed.
    """
    if mode not in ["pc", "verbose"]:
      raise ValueError("Mode must be 'pc' or 'verbose'")
    mapping = {
      "pc": 0,
      "verbose": 1,
    }
    await self.send_command(f"mode {mapping[mode]}")

  async def get_monitor_speed(self) -> int:
    """Get the global system (monitor) speed.

    Returns:
      Current monitor speed as a percentage (1-100)
    """
    response = await self.send_command("mspeed")
    return int(response)

  async def set_monitor_speed(self, speed_percent: int) -> None:
    """Set the global system (monitor) speed.

    Args:
      speed_percent: Speed percentage between 1 and 100, where 100 means full speed.

    Raises:
      ValueError: If speed_percent is not between 1 and 100.
    """
    if not (1 <= speed_percent <= 100):
      raise ValueError("Speed percent must be between 1 and 100")
    await self.send_command(f"mspeed {speed_percent}")

  async def nop(self) -> None:
    """No operation command.

    Does nothing except return the standard reply. Can be used to see if the link
    is active or to check for exceptions.
    """
    await self.send_command("nop")

  async def get_payload(self) -> int:
    """Get the payload percent value for the current robot.

    Returns:
      Current payload as a percentage of maximum (0-100)
    """
    response = await self.send_command("payload")
    return int(response)

  async def set_payload(self, payload_percent: int) -> None:
    """Set the payload percent of maximum for the currently selected or attached robot.

    Args:
      payload_percent: Payload percentage from 0 to 100 indicating the percent of the maximum payload the robot is carrying.

    Raises:
      ValueError: If payload_percent is not between 0 and 100.

    Note:
      If the robot is moving, waits for the robot to stop before setting a value.
    """
    if not (0 <= payload_percent <= 100):
      raise ValueError("Payload percent must be between 0 and 100")
    await self.send_command(f"payload {payload_percent}")

  async def set_parameter(
    self,
    data_id: int,
    value,
    unit_number: Optional[int] = None,
    sub_unit: Optional[int] = None,
    array_index: Optional[int] = None,
  ) -> None:
    """Change a value in the controller's parameter database.

    Args:
      data_id: DataID of parameter.
      value: New parameter value. If string, will be quoted automatically.
      unit_number: Unit number, usually the robot number (1 - N_ROB).
      sub_unit: Sub-unit, usually 0.
      array_index: Array index.

    Note:
      Updated values are not saved in flash unless a save-to-flash operation
      is performed (see DataID 901).
    """
    if unit_number is not None and sub_unit is not None and array_index is not None:
      # 5 argument format
      if isinstance(value, str):
        await self.send_command(f'pc {data_id} {unit_number} {sub_unit} {array_index} "{value}"')
      else:
        await self.send_command(f"pc {data_id} {unit_number} {sub_unit} {array_index} {value}")
    else:
      # 2 argument format
      if isinstance(value, str):
        await self.send_command(f'pc {data_id} "{value}"')
      else:
        await self.send_command(f"pc {data_id} {value}")

  async def get_parameter(
    self,
    data_id: int,
    unit_number: Optional[int] = None,
    sub_unit: Optional[int] = None,
    array_index: Optional[int] = None,
  ) -> str:
    """Get the value of a numeric parameter database item.

    Args:
      data_id: DataID of parameter.
      unit_number: Unit number, usually the robot number (1-NROB).
      sub_unit: Sub-unit, usually 0.
      array_index: Array index.

    Returns:
      str: The numeric value of the specified database parameter.
    """
    if unit_number is not None:
      if sub_unit is not None:
        if array_index is not None:
          response = await self.send_command(f"pd {data_id} {unit_number} {sub_unit} {array_index}")
        else:
          response = await self.send_command(f"pd {data_id} {unit_number} {sub_unit}")
      else:
        response = await self.send_command(f"pd {data_id} {unit_number}")
    else:
      response = await self.send_command(f"pd {data_id}")
    return response

  async def reset(self, robot_number: int) -> None:
    """Reset the threads associated with the specified robot.

    Stops and restarts the threads for the specified robot. Any TCP/IP connections
    made by these threads are broken. This command can only be sent to the status thread.

    Args:
      robot_number: The number of the robot thread to reset, from 1 to N_ROB. Must not be zero.

    Raises:
      ValueError: If robot_number is zero or negative.
    """
    if robot_number <= 0:
      raise ValueError("Robot number must be greater than zero")
    await self.send_command(f"reset {robot_number}")

  async def get_selected_robot(self) -> int:
    """Get the number of the currently selected robot.

    Returns:
      The number of the currently selected robot.
    """
    response = await self.send_command("selectRobot")
    return int(response)

  async def select_robot(self, robot_number: int) -> None:
    """Change the robot associated with this communications link.

    Does not affect the operation or attachment state of the robot. The status thread
    may select any robot or 0. Except for the status thread, a robot may only be
    selected by one thread at a time.

    Args:
      robot_number: The new robot to be connected to this thread (1 to N_ROB) or 0 for none.
    """
    await self.send_command(f"selectRobot {robot_number}")

  async def get_signal(self, signal_number: int) -> int:
    """Get the value of the specified digital input or output signal.

    Args:
      signal_number: The number of the digital signal to get.

    Returns:
      The current signal value.
    """
    response = await self.send_command(f"sig {signal_number}")
    sig_id, sig_val = response.split()
    return int(sig_val)

  async def set_signal(self, signal_number: int, value: int) -> None:
    """Set the specified digital input or output signal.

    Args:
      signal_number: The number of the digital signal to set.
      value: The signal value to set. 0 = off, non-zero = on.
    """
    await self.send_command(f"sig {signal_number} {value}")

  async def get_system_state(self) -> int:
    """Get the global system state code.

    Returns:
      The global system state code. Please see documentation for DataID 234.
    """
    response = await self.send_command("sysState")
    return int(response)

  async def get_tool_transformation_values(self) -> tuple[float, float, float, float, float, float]:
    """Get the current tool transformation values.

    Returns:
      A tuple containing (X, Y, Z, yaw, pitch, roll) for the tool transformation.
    """
    data = await self.send_command("tool")
    # Remove "tool:" prefix if present
    if data.startswith("tool: "):
      data = data[6:]

    parts = data.split()
    if len(parts) != 6:
      raise PreciseFlexError(-1, "Unexpected response format from tool command.")

    x, y, z, yaw, pitch, roll = self._parse_xyz_response(parts)

    return (x, y, z, yaw, pitch, roll)

  async def set_tool_transformation_values(
    self, x: float, y: float, z: float, yaw: float, pitch: float, roll: float
  ) -> None:
    """Set the robot tool transformation.

    The robot must be attached to set the tool. Setting the tool pauses any robot motion in progress.

    Args:
      x: Tool X coordinate.
      y: Tool Y coordinate.
      z: Tool Z coordinate.
      yaw: Tool yaw rotation.
      pitch: Tool pitch rotation.
      roll: Tool roll rotation.
    """
    await self.send_command(f"tool {x} {y} {z} {yaw} {pitch} {roll}")

  async def get_version(self) -> str:
    """Get the current version of TCS and any installed plug-ins.

    Returns:
      str: The current version information.
    """
    return await self.send_command("version")

  # region LOCATION COMMANDS

  async def get_location_angles(self, location_index: int) -> tuple[int, int, Dict[int, float]]:
    """Get the angle values for the specified station index.

    Args:
      location_index: The station index, from 1 to N_LOC.

    Returns:
      A tuple containing (type_code, station_index, angles_dict)

    Raises:
      PreciseFlexError: If attempting to get angles from a Cartesian location.
    """
    data = await self.send_command(f"locAngles {location_index}")
    parts = data.split(" ")

    type_code = int(parts[0])
    if type_code != 1:
      raise PreciseFlexError(-1, "Location is not of angles type.")

    station_index = int(parts[1])
    angles = self._parse_angles_response(parts[2:])

    return (type_code, station_index, angles)

  async def set_joint_angles(
    self,
    location_index: int,
    joint_position: Dict[int, float],
  ) -> None:
    """Set joint angles for stored location, handling rail configuration."""
    if self._has_rail:
      await self.send_command(
        f"locAngles {location_index} "
        f"{joint_position[PFAxis.RAIL]} "
        f"{joint_position[PFAxis.BASE]} "
        f"{joint_position[PFAxis.SHOULDER]} "
        f"{joint_position[PFAxis.ELBOW]} "
        f"{joint_position[PFAxis.WRIST]} "
        f"{joint_position[PFAxis.GRIPPER]}"
      )
    else:
      await self.send_command(
        f"locAngles {location_index} "
        f"{joint_position[PFAxis.BASE]} "
        f"{joint_position[PFAxis.SHOULDER]} "
        f"{joint_position[PFAxis.ELBOW]} "
        f"{joint_position[PFAxis.WRIST]} "
        f"{joint_position[PFAxis.GRIPPER]}"
      )

  async def get_location_xyz(
    self, location_index: int
  ) -> tuple[int, int, float, float, float, float, float, float]:
    """Get the Cartesian position values for the specified station index.

    Args:
      location_index: The station index, from 1 to N_LOC.

    Returns:
      A tuple containing (type_code, station_index, X, Y, Z, yaw, pitch, roll)

    Raises:
      PreciseFlexError: If attempting to get Cartesian position from an angles type location.
    """
    data = await self.send_command(f"locXyz {location_index}")
    parts = data.split(" ")

    type_code = int(parts[0])
    if type_code != 0:
      raise PreciseFlexError(-1, "Location is not of Cartesian type.")

    if len(parts) != 8:
      raise PreciseFlexError(-1, "Unexpected response format from locXyz command.")

    station_index = int(parts[1])
    x, y, z, yaw, pitch, roll = self._parse_xyz_response(parts[2:8])

    return (type_code, station_index, x, y, z, yaw, pitch, roll)

  async def set_location_xyz(
    self,
    location_index: int,
    cartesian_position: PreciseFlexCartesianCoords,
  ) -> None:
    """Set the Cartesian position values for the specified station index.

    Args:
      location_index: The station index, from 1 to N_LOC.
      cartesian_position: The Cartesian position to set.
    """
    await self.send_command(
      f"locXyz {location_index} "
      f"{cartesian_position.location.x} "
      f"{cartesian_position.location.y} "
      f"{cartesian_position.location.z} "
      f"{cartesian_position.rotation.yaw} "
      f"{cartesian_position.rotation.pitch} "
      f"{cartesian_position.rotation.roll}"
    )

  async def get_location_z_clearance(self, location_index: int) -> tuple[int, float, bool]:
    """Get the ZClearance and ZWorld properties for the specified location.

    Args:
      location_index: The station index, from 1 to N_LOC.

    Returns:
      A tuple containing (station_index, z_clearance, z_world)
    """
    data = await self.send_command(f"locZClearance {location_index}")
    parts = data.split(" ")

    if len(parts) != 3:
      raise PreciseFlexError(-1, "Unexpected response format from locZClearance command.")

    station_index = int(parts[0])
    z_clearance = float(parts[1])
    z_world = True if float(parts[2]) != 0 else False

    return (station_index, z_clearance, z_world)

  async def set_location_z_clearance(
    self, location_index: int, z_clearance: float, z_world: Optional[bool] = None
  ) -> None:
    """Set the ZClearance and ZWorld properties for the specified location.

    Args:
      location_index: The station index, from 1 to N_LOC.
      z_clearance: The new ZClearance property value.
      z_world (float, optional): The new ZWorld property value. If omitted, only ZClearance is set.
    """
    if z_world is None:
      await self.send_command(f"locZClearance {location_index} {z_clearance}")
    else:
      z_world_int = 1 if z_world else 0
      await self.send_command(f"locZClearance {location_index} {z_clearance} {z_world_int}")

  async def get_location_config(self, location_index: int) -> tuple[int, int]:
    """Get the Config property for the specified location.

    Args:
      location_index: The station index, from 1 to N_LOC.

    Returns:
      A tuple containing (station_index, config_value)
      config_value is a bit mask where:
      - 0 = None (no configuration specified)
      - 0x01 = GPL_Righty (right shouldered configuration)
      - 0x02 = GPL_Lefty (left shouldered configuration)
      - 0x04 = GPL_Above (elbow above the wrist)
      - 0x08 = GPL_Below (elbow below the wrist)
      - 0x10 = GPL_Flip (wrist pitched up)
      - 0x20 = GPL_NoFlip (wrist pitched down)
      - 0x1000 = GPL_Single (restrict wrist axis to +/- 180 degrees)
      Values can be combined using bitwise OR.
    """
    data = await self.send_command(f"locConfig {location_index}")
    parts = data.split(" ")

    if len(parts) != 2:
      raise PreciseFlexError(-1, "Unexpected response format from locConfig command.")

    station_index = int(parts[0])
    config_value = int(parts[1])

    return (station_index, config_value)

  async def set_location_config(self, location_index: int, config_value: int) -> None:
    """Set the Config property for the specified location.

    Args:
      location_index: The station index, from 1 to N_LOC.
      config_value: The new Config property value as a bit mask where:
      - 0 = None (no configuration specified)
      - 0x01 = GPL_Righty (right shouldered configuration)
      - 0x02 = GPL_Lefty (left shouldered configuration)
      - 0x04 = GPL_Above (elbow above the wrist)
      - 0x08 = GPL_Below (elbow below the wrist)
      - 0x10 = GPL_Flip (wrist pitched up)
      - 0x20 = GPL_NoFlip (wrist pitched down)
      - 0x1000 = GPL_Single (restrict wrist axis to +/- 180 degrees)
      Values can be combined using bitwise OR.

    Raises:
      ValueError: If config_value contains invalid bits or conflicting configurations.
    """
    # Define valid bit masks
    GPL_RIGHTY = 0x01
    GPL_LEFTY = 0x02
    GPL_ABOVE = 0x04
    GPL_BELOW = 0x08
    GPL_FLIP = 0x10
    GPL_NOFLIP = 0x20
    GPL_SINGLE = 0x1000

    # All valid bits
    ALL_VALID_BITS = (
      GPL_RIGHTY | GPL_LEFTY | GPL_ABOVE | GPL_BELOW | GPL_FLIP | GPL_NOFLIP | GPL_SINGLE
    )

    # Check for invalid bits
    if config_value & ~ALL_VALID_BITS:
      raise ValueError(f"Invalid config bits specified: 0x{config_value:X}")

    # Check for conflicting configurations
    if (config_value & GPL_RIGHTY) and (config_value & GPL_LEFTY):
      raise ValueError("Cannot specify both GPL_Righty and GPL_Lefty")

    if (config_value & GPL_ABOVE) and (config_value & GPL_BELOW):
      raise ValueError("Cannot specify both GPL_Above and GPL_Below")

    if (config_value & GPL_FLIP) and (config_value & GPL_NOFLIP):
      raise ValueError("Cannot specify both GPL_Flip and GPL_NoFlip")

    await self.send_command(f"locConfig {location_index} {config_value}")

  async def dest_c(self, arg1: int = 0) -> tuple[float, float, float, float, float, float, int]:
    """Get the destination or current Cartesian location of the robot.

    Args:
      arg1: Selects return value. Defaults to 0.
      0 = Return current Cartesian location if robot is not moving
      1 = Return target Cartesian location of the previous or current move

    Returns:
      A tuple containing (X, Y, Z, yaw, pitch, roll, config)
      If arg1 = 1 or robot is moving, returns the target location.
      If arg1 = 0 and robot is not moving, returns the current location.
    """
    if arg1 == 0:
      data = await self.send_command("destC")
    else:
      data = await self.send_command(f"destC {arg1}")

    parts = data.split()
    if len(parts) != 7:
      raise PreciseFlexError(-1, "Unexpected response format from destC command.")

    x, y, z, yaw, pitch, roll = self._parse_xyz_response(parts[:6])
    config = int(parts[6])

    return (x, y, z, yaw, pitch, roll, config)

  async def dest_j(self, arg1: int = 0) -> Dict[int, float]:
    """Get the destination or current joint location of the robot.

    Args:
      arg1: Selects return value. Defaults to 0.
      0 = Return current joint location if robot is not moving
      1 = Return target joint location of the previous or current move

    Returns:
      A dict mapping PFAxis to float values.
      If arg1 = 1 or robot is moving, returns the target joint positions.
      If arg1 = 0 and robot is not moving, returns the current joint positions.
    """
    if arg1 == 0:
      data = await self.send_command("destJ")
    else:
      data = await self.send_command(f"destJ {arg1}")

    parts = data.split()
    if not parts:
      raise PreciseFlexError(-1, "Unexpected response format from destJ command.")

    return self._parse_angles_response(parts)

  async def here_j(self, location_index: int) -> None:
    """Record the current position of the selected robot into the specified Location as angles.

    The Location is automatically set to type "angles".

    Args:
      location_index: The station index, from 1 to N_LOC.
    """
    await self.send_command(f"hereJ {location_index}")

  async def here_c(self, location_index: int) -> None:
    """Record the current position of the selected robot into the specified Location as Cartesian.

    The Location object is automatically set to type "Cartesian".
    Can be used to change the pallet origin (index 1,1,1) value.

    Args:
      location_index: The station index, from 1 to N_LOC.
    """
    await self.send_command(f"hereC {location_index}")

  # region PROFILE COMMANDS

  async def get_profile_speed(self, profile_index: int) -> float:
    """Get the speed property of the specified profile.

    Args:
      profile_index: The profile index to query.

    Returns:
      float: The current speed as a percentage. 100 = full speed.
    """
    response = await self.send_command(f"Speed {profile_index}")
    profile, speed = response.split()
    return float(speed)

  async def set_profile_speed(self, profile_index: int, speed_percent: float) -> None:
    """Set the speed property of the specified profile.

    Args:
      profile_index: The profile index to modify.
      speed_percent: The new speed as a percentage. 100 = full speed.
      Values > 100 may be accepted depending on system configuration.
    """
    await self.send_command(f"Speed {profile_index} {speed_percent}")

  async def get_profile_speed2(self, profile_index: int) -> float:
    """Get the speed2 property of the specified profile.

    Args:
      profile_index: The profile index to query.

    Returns:
      float: The current speed2 as a percentage. Used for Cartesian moves.
    """
    response = await self.send_command(f"Speed2 {profile_index}")
    profile, speed2 = response.split()
    return float(speed2)

  async def set_profile_speed2(self, profile_index: int, speed2_percent: float) -> None:
    """Set the speed2 property of the specified profile.

    Args:
      profile_index: The profile index to modify.
      speed2_percent: The new speed2 as a percentage. 100 = full speed.
      Used for Cartesian moves. Normally set to 0.
    """
    await self.send_command(f"Speed2 {profile_index} {speed2_percent}")

  async def get_profile_accel(self, profile_index: int) -> float:
    """Get the acceleration property of the specified profile.

    Args:
      profile_index: The profile index to query.

    Returns:
      float: The current acceleration as a percentage. 100 = maximum acceleration.
    """
    response = await self.send_command(f"Accel {profile_index}")
    profile, accel = response.split()
    return float(accel)

  async def set_profile_accel(self, profile_index: int, accel_percent: float) -> None:
    """Set the acceleration property of the specified profile.

    Args:
      profile_index: The profile index to modify.
      accel_percent: The new acceleration as a percentage. 100 = maximum acceleration.
      Maximum value depends on system configuration.
    """
    await self.send_command(f"Accel {profile_index} {accel_percent}")

  async def get_profile_accel_ramp(self, profile_index: int) -> float:
    """Get the acceleration ramp property of the specified profile.

    Args:
      profile_index: The profile index to query.

    Returns:
      float: The current acceleration ramp time in seconds.
    """
    response = await self.send_command(f"AccRamp {profile_index}")
    profile, accel_ramp = response.split()
    return float(accel_ramp)

  async def set_profile_accel_ramp(self, profile_index: int, accel_ramp_seconds: float) -> None:
    """Set the acceleration ramp property of the specified profile.

    Args:
      profile_index: The profile index to modify.
      accel_ramp_seconds: The new acceleration ramp time in seconds.
    """
    await self.send_command(f"AccRamp {profile_index} {accel_ramp_seconds}")

  async def get_profile_decel(self, profile_index: int) -> float:
    """Get the deceleration property of the specified profile.

    Args:
      profile_index: The profile index to query.

    Returns:
      float: The current deceleration as a percentage. 100 = maximum deceleration.
    """
    response = await self.send_command(f"Decel {profile_index}")
    profile, decel = response.split()
    return float(decel)

  async def set_profile_decel(self, profile_index: int, decel_percent: float) -> None:
    """Set the deceleration property of the specified profile.

    Args:
      profile_index: The profile index to modify.
      decel_percent: The new deceleration as a percentage. 100 = maximum deceleration.
      Maximum value depends on system configuration.
    """
    await self.send_command(f"Decel {profile_index} {decel_percent}")

  async def get_profile_decel_ramp(self, profile_index: int) -> float:
    """Get the deceleration ramp property of the specified profile.

    Args:
      profile_index: The profile index to query.

    Returns:
      float: The current deceleration ramp time in seconds.
    """
    response = await self.send_command(f"DecRamp {profile_index}")
    profile, decel_ramp = response.split()
    return float(decel_ramp)

  async def set_profile_decel_ramp(self, profile_index: int, decel_ramp_seconds: float) -> None:
    """Set the deceleration ramp property of the specified profile.

    Args:
      profile_index: The profile index to modify.
      decel_ramp_seconds: The new deceleration ramp time in seconds.
    """
    await self.send_command(f"DecRamp {profile_index} {decel_ramp_seconds}")

  async def get_profile_in_range(self, profile_index: int) -> float:
    """Get the InRange property of the specified profile.

    Args:
      profile_index: The profile index to query.

    Returns:
      float: The current InRange value (-1 to 100).
      -1 = do not stop at end of motion if blending is possible
      0 = always stop but do not check end point error
      > 0 = wait until close to end point (larger numbers mean less position error allowed)
    """
    response = await self.send_command(f"InRange {profile_index}")
    profile, in_range = response.split()
    return float(in_range)

  async def set_profile_in_range(self, profile_index: int, in_range_value: float) -> None:
    """Set the InRange property of the specified profile.

    Args:
      profile_index: The profile index to modify.
      in_range_value: The new InRange value from -1 to 100.
      -1 = do not stop at end of motion if blending is possible
      0 = always stop but do not check end point error
      > 0 = wait until close to end point (larger numbers mean less position error allowed)

    Raises:
      ValueError: If in_range_value is not between -1 and 100.
    """
    if not (-1 <= in_range_value <= 100):
      raise ValueError("InRange value must be between -1 and 100")
    await self.send_command(f"InRange {profile_index} {in_range_value}")

  async def get_profile_straight(self, profile_index: int) -> bool:
    """Get the Straight property of the specified profile.

    Args:
      profile_index: The profile index to query.

    Returns:
      The current Straight property value.
      True = follow a straight-line path
      False = follow a joint-based path (coordinated axes movement)
    """
    response = await self.send_command(f"Straight {profile_index}")
    profile, straight = response.split()

    return straight == "True"

  async def set_profile_straight(self, profile_index: int, straight_mode: bool) -> None:
    """Set the Straight property of the specified profile.

    Args:
      profile_index: The profile index to modify.
      straight_mode: The path type to use.
      True = follow a straight-line path
      False = follow a joint-based path (robot axes move in coordinated manner)

    Raises:
      ValueError: If straight_mode is not True or False.
    """
    straight_int = 1 if straight_mode else 0
    await self.send_command(f"Straight {profile_index} {straight_int}")

  async def set_motion_profile_values(
    self,
    profile: int,
    speed: float,
    speed2: float,
    acceleration: float,
    deceleration: float,
    acceleration_ramp: float,
    deceleration_ramp: float,
    in_range: float,
    straight: bool,
  ):
    """
    Set motion profile values for the specified profile index on the PreciseFlex robot.

    Args:
      profile: Profile index to set values for.
      speed: Percentage of maximum speed. 100 = full speed. Values >100 may be accepted depending on system config.
      speed2: Secondary speed setting, typically for Cartesian moves. Normally 0. Interpreted as a percentage.
      acceleration: Percentage of maximum acceleration. 100 = full accel.
      deceleration: Percentage of maximum deceleration. 100 = full decel.
      acceleration_ramp: Acceleration ramp time in seconds.
      deceleration_ramp: Deceleration ramp time in seconds.
      in_range: InRange value, from -1 to 100. -1 = allow blending, 0 = stop without checking, >0 = enforce position accuracy.
      straight: If True, follow a straight-line path (-1). If False, follow a joint-based path (0).
    """
    if not (0 <= speed):
      raise ValueError("Speed must be > 0 (percent).")

    if not (0 <= speed2):
      raise ValueError("Speed2 must be > 0 (percent).")

    if not (0 <= acceleration <= 100):
      raise ValueError("Acceleration must be between 0 and 100 (percent).")

    if not (0 <= deceleration <= 100):
      raise ValueError("Deceleration must be between 0 and 100 (percent).")

    if acceleration_ramp < 0:
      raise ValueError("Acceleration ramp must be >= 0 (seconds).")

    if deceleration_ramp < 0:
      raise ValueError("Deceleration ramp must be >= 0 (seconds).")

    if not (-1 <= in_range <= 100):
      raise ValueError("InRange must be between -1 and 100.")

    straight_int = -1 if straight else 0
    await self.send_command(
      f"Profile {profile} {speed} {speed2} {acceleration} {deceleration} {acceleration_ramp} {deceleration_ramp} {in_range} {straight_int}"
    )

  async def get_motion_profile_values(
    self, profile: int
  ) -> tuple[int, float, float, float, float, float, float, float, bool]:
    """
    Get the current motion profile values for the specified profile index on the PreciseFlex robot.

    Args:
      profile: Profile index to get values for.

    Returns:
      A tuple containing (profile, speed, speed2, acceleration, deceleration, acceleration_ramp, deceleration_ramp, in_range, straight)
        - profile: Profile index
        - speed: Percentage of maximum speed
        - speed2: Secondary speed setting
        - acceleration: Percentage of maximum acceleration
        - deceleration: Percentage of maximum deceleration
        - acceleration_ramp: Acceleration ramp time in seconds
        - deceleration_ramp: Deceleration ramp time in seconds
        - in_range: InRange value (-1 to 100)
        - straight: True if straight-line path, False if joint-based path
    """
    data = await self.send_command(f"Profile {profile}")
    parts = data.split(" ")
    if len(parts) != 9:
      raise PreciseFlexError(-1, "Unexpected response format from device.")

    return (
      int(parts[0]),
      float(parts[1]),
      float(parts[2]),
      float(parts[3]),
      float(parts[4]),
      float(parts[5]),
      float(parts[6]),
      float(parts[7]),
      False if int(parts[8]) == 0 else True,
    )

  # region MOTION COMMANDS
  async def move_to_stored_location(self, location_index: int, profile_index: int) -> None:
    """Move to the location specified by the station index using the specified profile.

    Args:
      location_index: The index of the location to which the robot moves.
      profile_index: The profile index for this move.

    Note:
      Requires that the robot be attached.
    """
    await self.send_command(f"move {location_index} {profile_index}")

  async def move_to_stored_location_appro(self, location_index: int, profile_index: int) -> None:
    """Approach the location specified by the station index using the specified profile.

    This is similar to `move_to_stored_location` except that the Z clearance value is included.

    Args:
      location_index: The index of the location to which the robot moves.
      profile_index: The profile index for this move.

    Note:
      Requires that the robot be attached.
    """
    await self.send_command(f"moveAppro {location_index} {profile_index}")

  async def move_extra_axis(
    self, axis1_position: float, axis2_position: Optional[float] = None
  ) -> None:
    """Post a move for one or two extra axes during the next Cartesian motion.

    Does not cause the robot to move at this time. Only some kinematic modules support extra axes.

    Args:
      axis1_position: The destination position for the 1st extra axis.
      axis2_position (float, optional): The destination position for the 2nd extra axis, if any.

    Note:
      Requires that the robot be attached.
    """
    if axis2_position is None:
      await self.send_command(f"moveExtraAxis {axis1_position}")
    else:
      await self.send_command(f"moveExtraAxis {axis1_position} {axis2_position}")

  async def move_one_axis(
    self, axis_number: int, destination_position: float, profile_index: int
  ) -> None:
    """Move a single axis to the specified position using the specified profile.

    Args:
      axis_number: The number of the axis to move.
      destination_position: The destination position for this axis.
      profile_index: The index of the profile to use during this motion.

    Note:
      Requires that the robot be attached.
    """
    await self.send_command(f"moveOneAxis {axis_number} {destination_position} {profile_index}")

  async def move_c(
    self,
    profile_index: int,
    cartesian_coords: PreciseFlexCartesianCoords,
  ) -> None:
    """Move the robot to the Cartesian location specified by the arguments.

    Args:
      profile_index: The profile index to use for this motion.
      cartesian_coords: The Cartesian coordinates to which the robot should move.

    Note:
      Requires that the robot be attached.
    """

    cmd = (
      f"moveC {profile_index} "
      f"{cartesian_coords.location.x} "
      f"{cartesian_coords.location.y} "
      f"{cartesian_coords.location.z} "
      f"{cartesian_coords.rotation.yaw} "
      f"{cartesian_coords.rotation.pitch} "
      f"{cartesian_coords.rotation.roll} "
    )

    if cartesian_coords.orientation is not None:
      config_int = self._convert_orientation_enum_to_int(cartesian_coords.orientation)
      config_int |= 0x1000  # GPL_Single: restrict wrist to ±180°
      cmd += f"{config_int}"

    await self.send_command(cmd)

  async def move_j(self, profile_index: int, joint_coords: Dict[int, float]) -> None:
    """Move the robot using joint coordinates, handling rail configuration."""
    if self._has_rail:
      angles_str = (
        f"{joint_coords[PFAxis.BASE]} "
        f"{joint_coords[PFAxis.SHOULDER]} "
        f"{joint_coords[PFAxis.ELBOW]} "
        f"{joint_coords[PFAxis.WRIST]} "
        f"{joint_coords[PFAxis.GRIPPER]} "
        f"{joint_coords[PFAxis.RAIL]} "
      )
    else:
      angles_str = (
        f"{joint_coords[PFAxis.BASE]} "
        f"{joint_coords[PFAxis.SHOULDER]} "
        f"{joint_coords[PFAxis.ELBOW]} "
        f"{joint_coords[PFAxis.WRIST]} "
        f"{joint_coords[PFAxis.GRIPPER]}"
      )
    await self.send_command(f"moveJ {profile_index} {angles_str}")

  async def release_brake(self, axis: int) -> None:
    """Release the axis brake.

    Overrides the normal operation of the brake. It is important that the brake not be set
    while a motion is being performed. This feature is used to lock an axis to prevent
    motion or jitter.

    Args:
      axis: The number of the axis whose brake should be released.
    """
    await self.send_command(f"releaseBrake {axis}")

  async def set_brake(self, axis: int) -> None:
    """Set the axis brake.

    Overrides the normal operation of the brake. It is important not to set a brake on an
    axis that is moving as it may damage the brake or damage the motor.

    Args:
      axis: The number of the axis whose brake should be set.
    """
    await self.send_command(f"setBrake {axis}")

  async def state(self) -> str:
    """Return state of motion.

    This value indicates the state of the currently executing or last completed robot motion.
    For additional information, please see 'Robot.TrajState' in the GPL reference manual.

    Returns:
      str: The current motion state.
    """
    return await self.send_command("state")

  async def wait_for_eom(self) -> None:
    """Wait for the robot to reach the end of the current motion.

    Waits for the robot to reach the end of the current motion or until it is stopped by
    some other means. Does not reply until the robot has stopped.
    """
    await self.send_command("waitForEom")
    await asyncio.sleep(0.2)  # Small delay to ensure command is fully processed

  async def zero_torque(self, enable: bool, axis_mask: int = 1) -> None:
    """Sets or clears zero torque mode for the selected robot.

    Individual axes may be placed into zero torque mode while the remaining axes are servoing.

    Args:
      enable: If True, enable torque mode for axes specified by axis_mask.  If False, disable torque mode for the entire robot.
      axis_mask: The bit mask specifying the axes to be placed in torque mode when enable is True.  The mask is computed by OR'ing the axis bits: 1 = axis 1, 2 = axis 2, 4 = axis 3, 8 = axis 4, etc.  Ignored when enable is False.
    """

    if enable:
      assert axis_mask > 0, "axis_mask must be greater than 0"
      await self.send_command(f"zeroTorque 1 {axis_mask}")
    else:
      await self.send_command("zeroTorque 0")

  # region PAROBOT COMMANDS

  async def change_config(self, grip_mode: int = 0) -> None:
    """Change Robot configuration from Righty to Lefty or vice versa using customizable locations.

    Uses customizable locations to avoid hitting robot during change.
    Does not include checks for collision inside work volume of the robot.
    Can be customized by user for their work cell configuration.

    Args:
      grip_mode: Gripper control mode.
      0 = do not change gripper (default)
      1 = open gripper
      2 = close gripper
    """
    await self.send_command(f"ChangeConfig {grip_mode}")

  async def change_config2(self, grip_mode: int = 0) -> None:
    """Change Robot configuration from Righty to Lefty or vice versa using algorithm.

    Uses an algorithm to avoid hitting robot during change.
    Does not include checks for collision inside work volume of the robot.
    Can be customized by user for their work cell configuration.

    Args:
      grip_mode: Gripper control mode.
      0 = do not change gripper (default)
      1 = open gripper
      2 = close gripper
    """
    await self.send_command(f"ChangeConfig2 {grip_mode}")

  async def get_grasp_data(self) -> tuple[float, float, float]:
    """Get the data to be used for the next force-controlled PickPlate command grip operation.

    Returns:
      A tuple containing (plate_width_mm, finger_speed_percent, grasp_force)
    """
    data = await self.send_command("GraspData")
    parts = data.split()

    if len(parts) != 3:
      raise PreciseFlexError(-1, "Unexpected response format from GraspData command.")

    plate_width = float(parts[0])
    finger_speed = float(parts[1])
    grasp_force = float(parts[2])

    return (plate_width, finger_speed, grasp_force)

  async def set_grasp_data(
    self, plate_width: float, finger_speed_percent: float, grasp_force: float
  ) -> None:
    """Set the data to be used for the next force-controlled PickPlate command grip operation.

    This data remains in effect until the next GraspData command or the system is restarted.

    Args:
      plate_width_mm: The plate width in mm.
      finger_speed_percent: The finger speed during grasp where 100 means 100%.
      grasp_force: The gripper squeezing force, in Newtons.
      A positive value indicates the fingers must close to grasp.
      A negative value indicates the fingers must open to grasp.
    """
    await self.send_command(f"GraspData {plate_width} {finger_speed_percent} {grasp_force}")

  async def _get_grip_close_pos(self) -> float:
    """Get the gripper close position for the servoed gripper.

    Returns:
      float: The current gripper close position.
    """
    data = await self.send_command("GripClosePos")
    return float(data)

  async def _set_grip_close_pos(self, close_position: float) -> None:
    """Set the gripper close position for the servoed gripper.

    The close position may be changed by a force-controlled grip operation.

    Args:
      close_position: The new gripper close position.
    """
    await self.send_command(f"GripClosePos {close_position}")

  async def _get_grip_open_pos(self) -> float:
    """Get the gripper open position for the servoed gripper.

    Returns:
      float: The current gripper open position.
    """
    data = await self.send_command("GripOpenPos")
    return float(data)

  async def _set_grip_open_pos(self, open_position: float) -> None:
    """Set the gripper open position for the servoed gripper.

    Args:
      open_position: The new gripper open position.
    """
    await self.send_command(f"GripOpenPos {open_position}")

  async def move_rail(
    self, station_id: Optional[int] = None, mode: int = 0, rail_destination: Optional[float] = None
  ) -> None:
    """Moves the optional linear rail.

    The rail may be moved immediately or simultaneously with the next pick or place motion.
    The location may be associated with the station or specified explicitly.

    Args:
      station_id: The destination station ID. Only used if rail_destination is omitted.
      mode: Mode of operation.
      0 or omitted = cancel any pending MoveRail
      1 = Move rail immediately
      2 = Move rail during next pick or place
      rail_destination (float, optional): If specified, use this value as the rail destination
      rather than the station location.
    """
    if rail_destination is not None:
      await self.send_command(f"MoveRail {station_id or ''} {mode} {rail_destination}")
    elif station_id is not None:
      await self.send_command(f"MoveRail {station_id} {mode}")
    else:
      await self.send_command(f"MoveRail {mode}")

  async def get_pallet_index(self, station_id: int) -> tuple[int, int, int, int]:
    """Get the current pallet index values for the specified station.

    Args:
      station_id: Station ID, from 1 to N_LOC.

    Returns:
      A tuple containing (station_id, pallet_index_x, pallet_index_y, pallet_index_z)
    """
    data = await self.send_command(f"PalletIndex {station_id}")
    parts = data.split()

    if len(parts) != 4:
      raise PreciseFlexError(-1, "Unexpected response format from PalletIndex command.")

    station_id = int(parts[0])
    pallet_index_x = int(parts[1])
    pallet_index_y = int(parts[2])
    pallet_index_z = int(parts[3])

    return (station_id, pallet_index_x, pallet_index_y, pallet_index_z)

  async def set_pallet_index(
    self, station_id: int, pallet_index_x: int = 0, pallet_index_y: int = 0, pallet_index_z: int = 0
  ) -> None:
    """Set the pallet index value from 1 to n of the station used by subsequent pick or place.

    If an index argument is 0 or omitted, the corresponding index is not changed.
    Negative values generate an error.

    Args:
      station_id: Station ID, from 1 to N_LOC.
      pallet_index_x: Pallet index X. If 0 or omitted, X index is not changed.
      pallet_index_y: Pallet index Y. If 0 or omitted, Y index is not changed.
      pallet_index_z: Pallet index Z. If 0 or omitted, Z index is not changed.

    Raises:
      ValueError: If any index value is negative.
    """
    if pallet_index_x < 0:
      raise ValueError("Pallet index X cannot be negative")
    if pallet_index_y < 0:
      raise ValueError("Pallet index Y cannot be negative")
    if pallet_index_z < 0:
      raise ValueError("Pallet index Z cannot be negative")

    await self.send_command(
      f"PalletIndex {station_id} {pallet_index_x} {pallet_index_y} {pallet_index_z}"
    )

  async def get_pallet_origin(
    self, station_id: int
  ) -> tuple[int, float, float, float, float, float, float, int]:
    """Get the current pallet origin data for the specified station.

    Args:
      station_id: Station ID, from 1 to N_LOC.

    Returns:
      A tuple containing (station_id, x, y, z, yaw, pitch, roll, config)
    """
    data = await self.send_command(f"PalletOrigin {station_id}")
    parts = data.split()

    if len(parts) != 8:
      raise PreciseFlexError(-1, "Unexpected response format from PalletOrigin command.")

    station_id = int(parts[0])
    x = float(parts[1])
    y = float(parts[2])
    z = float(parts[3])
    yaw = float(parts[4])
    pitch = float(parts[5])
    roll = float(parts[6])
    config = int(parts[7])

    return (station_id, x, y, z, yaw, pitch, roll, config)

  async def set_pallet_origin(
    self,
    station_id: int,
    cartesian_coords: PreciseFlexCartesianCoords,
  ) -> None:
    """Define the origin of a pallet reference frame.

    Specifies the world location and orientation of the (1,1,1) pallet position.
    Must be followed by a PalletX command.

    The orientation and configuration specified here determines the world orientation
    of the robot during all pick or place operations using this pallet.

    Args:
      station_id: Station ID, from 1 to N_LOC.
      cartesian_coords: The Cartesian coordinates defining the pallet origin.
    """

    cmd = (
      f"PalletOrigin {station_id} "
      f"{cartesian_coords.location.x} "
      f"{cartesian_coords.location.y} "
      f"{cartesian_coords.location.z} "
      f"{cartesian_coords.rotation.yaw} "
      f"{cartesian_coords.rotation.pitch} "
      f"{cartesian_coords.rotation.roll} "
    )

    if cartesian_coords.orientation is not None:
      config_int = self._convert_orientation_enum_to_int(cartesian_coords.orientation)
      cmd += f"{config_int}"

    await self.send_command(cmd)

  async def get_pallet_x(self, station_id: int) -> tuple[int, int, float, float, float]:
    """Get the current pallet X data for the specified station.

    Args:
      station_id: Station ID, from 1 to N_LOC.

    Returns:
      A tuple containing (station_id, x_position_count, world_x, world_y, world_z)
    """
    data = await self.send_command(f"PalletX {station_id}")
    parts = data.split()

    if len(parts) != 5:
      raise PreciseFlexError(-1, "Unexpected response format from PalletX command.")

    station_id = int(parts[0])
    x_position_count = int(parts[1])
    world_x = float(parts[2])
    world_y = float(parts[3])
    world_z = float(parts[4])

    return (station_id, x_position_count, world_x, world_y, world_z)

  async def set_pallet_x(
    self, station_id: int, x_position_count: int, world_x: float, world_y: float, world_z: float
  ) -> None:
    """Define the last point on the pallet X axis.

    Specifies the world location of the (n,1,1) pallet position, where n is the x_position_count value.
    Must follow a PalletOrigin command.

    Args:
      station_id: Station ID, from 1 to N_LOC.
      x_position_count: X position count.
      world_x: World location X coordinate.
      world_y: World location Y coordinate.
      world_z: World location Z coordinate.
    """
    await self.send_command(
      f"PalletX {station_id} {x_position_count} {world_x} {world_y} {world_z}"
    )

  async def get_pallet_y(self, station_id: int) -> tuple[int, int, float, float, float]:
    """Get the current pallet Y data for the specified station.

    Args:
      station_id: Station ID, from 1 to N_LOC.

    Returns:
      A tuple containing (station_id, y_position_count, world_x, world_y, world_z)
    """
    data = await self.send_command(f"PalletY {station_id}")
    parts = data.split()

    if len(parts) != 5:
      raise PreciseFlexError(-1, "Unexpected response format from PalletY command.")

    station_id = int(parts[0])
    y_position_count = int(parts[1])
    world_x = float(parts[2])
    world_y = float(parts[3])
    world_z = float(parts[4])

    return (station_id, y_position_count, world_x, world_y, world_z)

  async def set_pallet_y(
    self, station_id: int, y_position_count: int, world_x: float, world_y: float, world_z: float
  ) -> None:
    """Define the last point on the pallet Y axis.

    Specifies the world location of the (1,n,1) pallet position, where n is the y_position_count value.
    If this command is executed, a 2 or 3-dimensional pallet is assumed.
    Must follow a PalletX command.

    Args:
      station_id: Station ID, from 1 to N_LOC.
      y_position_count: Y position count.
      world_x: World location X coordinate.
      world_y: World location Y coordinate.
      world_z: World location Z coordinate.
    """
    await self.send_command(
      f"PalletY {station_id} {y_position_count} {world_x} {world_y} {world_z}"
    )

  async def get_pallet_z(self, station_id: int) -> tuple[int, int, float, float, float]:
    """Get the current pallet Z data for the specified station.

    Args:
      station_id: Station ID, from 1 to N_LOC.

    Returns:
      A tuple containing (station_id, z_position_count, world_x, world_y, world_z)
    """
    data = await self.send_command(f"PalletZ {station_id}")
    parts = data.split()

    if len(parts) != 5:
      raise PreciseFlexError(-1, "Unexpected response format from PalletZ command.")

    station_id = int(parts[0])
    z_position_count = int(parts[1])
    world_x = float(parts[2])
    world_y = float(parts[3])
    world_z = float(parts[4])

    return (station_id, z_position_count, world_x, world_y, world_z)

  async def set_pallet_z(
    self, station_id: int, z_position_count: int, world_x: float, world_y: float, world_z: float
  ) -> None:
    """Define the last point on the pallet Z axis.

    Specifies the world location of the (1,1,n) pallet position, where n is the z_position_count value.
    If this command is executed, a 3-dimensional pallet is assumed.
    Must follow a PalletX and PalletY command.

    Args:
      station_id: Station ID, from 1 to N_LOC.
      z_position_count: Z position count.
      world_x: World location X coordinate.
      world_y: World location Y coordinate.
      world_z: World location Z coordinate.
    """
    await self.send_command(
      f"PalletZ {station_id} {z_position_count} {world_x} {world_y} {world_z}"
    )

  async def pick_plate_station(
    self,
    station_id: int,
    horizontal_compliance: bool = False,
    horizontal_compliance_torque: int = 0,
  ) -> bool:
    """Moves to a predefined position or pallet location and picks up plate.

    If the arm must change configuration, it automatically goes through the Park position.
    At the conclusion of this routine, the arm is left gripping the plate and stopped at the nest approach position.
    Use Teach function to teach station pick point.

    Args:
      station_id: Station ID, from 1 to Max.
      horizontal_compliance: If True, enable horizontal compliance while closing the gripper to allow centering around the plate.
      horizontal_compliance_torque: The % of the original horizontal holding torque to be retained during compliance. If omitted, 0 is used.

    Returns:
      bool: True if the plate was successfully grasped or force control was not used.
        False if the force-controlled gripper detected no plate present.
    """
    horizontal_compliance_int = 1 if horizontal_compliance else 0
    ret_code = await self.send_command(
      f"PickPlate {station_id} {horizontal_compliance_int} {horizontal_compliance_torque}"
    )
    return ret_code != "0"

  async def place_plate_station(
    self,
    station_id: int,
    horizontal_compliance: bool = False,
    horizontal_compliance_torque: int = 0,
  ) -> None:
    """Moves to a predefined position or pallet location and places a plate.

    If the arm must change configuration, it automatically goes through the Park position.
    At the conclusion of this routine, the arm is left gripping the plate and stopped at the nest approach position.
    Use Teach function to teach station place point.

    Args:
    station_id: Station ID, from 1 to Max.
    horizontal_compliance: If True, enable horizontal compliance during the move to place the plate, to allow centering in the fixture.
    horizontal_compliance_torque: The % of the original horizontal holding torque to be retained during compliance. If omitted, 0 is used.
    """
    horizontal_compliance_int = 1 if horizontal_compliance else 0
    await self.send_command(
      f"PlacePlate {station_id} {horizontal_compliance_int} {horizontal_compliance_torque}"
    )

  async def get_rail_position(self, station_id: int) -> float:
    """Get the position of the optional rail axis that is associated with a station.

    Args:
    station_id: Station ID, from 1 to Max.

    Returns:
    float: The current rail position for the specified station.
    """
    data = await self.send_command(f"Rail {station_id}")
    return float(data)

  async def set_rail_position(self, station_id: int, rail_position: float) -> None:
    """Set the position of the optional rail axis that is associated with a station.

    The station rail data is loaded and saved by the LoadFile and StoreFile commands.

    Args:
    station_id: Station ID, from 1 to Max.
    rail_position: The new rail position.
    """
    await self.send_command(f"Rail {station_id} {rail_position}")

  async def teach_plate_station(self, station_id: int, z_clearance: float = 50.0) -> None:
    """Sets the plate location to the current robot position and configuration.

    The location is saved as Cartesian coordinates. Z clearance must be high enough to withdraw the gripper.
    If this station is a pallet, the pallet indices must be set to 1, 1, 1. The pallet frame is not changed,
    only the location relative to the pallet.

    Args:
    station_id: Station ID, from 1 to Max.
    z_clearance: The Z Clearance value. If omitted, a value of 50 is used. If specified and non-zero, this value is used.
    """
    await self.send_command(f"TeachPlate {station_id} {z_clearance}")

  async def get_station_type(self, station_id: int) -> tuple[int, int, int, float, float, float]:
    """Get the station configuration for the specified station ID.

    Args:
      station_id: Station ID, from 1 to Max.

    Returns:
      A tuple containing (station_id, access_type, location_type, z_clearance, z_above, z_grasp_offset)
      - station_id: The station ID
      - access_type: 0 = horizontal, 1 = vertical
      - location_type: 0 = normal single, 1 = pallet (1D, 2D, 3D)
      - z_clearance: ZClearance value in mm
      - z_above: ZAbove value in mm
      - z_grasp_offset: ZGrasp offset
    """
    data = await self.send_command(f"StationType {station_id}")
    parts = data.split()

    if len(parts) != 6:
      raise PreciseFlexError(-1, "Unexpected response format from StationType command.")

    station_id = int(parts[0])
    access_type = int(parts[1])
    location_type = int(parts[2])
    z_clearance = float(parts[3])
    z_above = float(parts[4])
    z_grasp_offset = float(parts[5])

    return (station_id, access_type, location_type, z_clearance, z_above, z_grasp_offset)

  async def set_station_type(
    self,
    station_id: int,
    access_type: int,
    location_type: int,
    z_clearance: float,
    z_above: float,
    z_grasp_offset: float,
  ) -> None:
    """Set the station configuration for the specified station ID.

    Args:
      station_id: Station ID, from 1 to Max.
      access_type: The station access type.
      0 = horizontal (for "hotel" carriers accessed by horizontal move)
      1 = vertical (for stacks or tube racks accessed with vertical motion)
      location_type: The location type.
      0 = normal single location
      1 = pallet (1D, 2D, or 3D regular arrays requiring column, row, and layer index)
      z_clearance: ZClearance value in mm. The horizontal or vertical distance
      from the final location used when approaching or departing from a station.
      z_above: ZAbove value in mm. The vertical offset used with horizontal
      access when approaching or departing from the location.
      z_grasp_offset: ZGrasp offset. Added to ZClearance when an object is
      being held to compensate for the part in the gripper.

    Raises:
      ValueError: If access_type or location_type are not valid values.
    """
    if access_type not in [0, 1]:
      raise ValueError("Access type must be 0 (horizontal) or 1 (vertical)")

    if location_type not in [0, 1]:
      raise ValueError("Location type must be 0 (normal single) or 1 (pallet)")

    await self.send_command(
      f"StationType {station_id} {access_type} {location_type} {z_clearance} {z_above} {z_grasp_offset}"
    )

  # region SSGRIP COMMANDS

  async def home_all_if_no_plate(self) -> int:
    """Tests if the gripper is holding a plate. If not, enable robot power and home all robots.

    Returns:
      -1 if no plate detected and the command succeeded, 0 if a plate was detected.
    """
    response = await self.send_command("HomeAll_IfNoPlate")
    return int(response)

  async def _grasp_plate(
    self, plate_width_mm: float, finger_speed_percent: int, grasp_force: float
  ) -> int:
    """Grasps a plate with limited force.

    Low level method. Use `pick_plate` instead for typical pick-and-place operations.

    A plate can be grasped by opening or closing the gripper. The actual commanded gripper
    width generated by this function is a few mm smaller (or larger) than plate_width_mm
    to permit the servos PID loop to generate the gripping force.

    Args:
      plate_width_mm: Plate width in mm. Should be accurate to within about 1 mm.
      finger_speed_percent: Percent speed to close fingers. 1 to 100.
      grasp_force: Maximum gripper squeeze force in Newtons.
        A positive value indicates the fingers must close to grasp.
        A negative value indicates the fingers must open to grasp.

    Returns:
      -1 if the plate has been grasped, 0 if the final gripping force indicates no plate.

    Raises:
      ValueError: If finger_speed_percent is not between 1 and 100.
    """
    if not (1 <= finger_speed_percent <= 100):
      raise ValueError("Finger speed percent must be between 1 and 100")

    response = await self.send_command(
      f"GraspPlate {plate_width_mm} {finger_speed_percent} {grasp_force}"
    )
    return int(response)

  async def _release_plate(
    self, open_width_mm: float, finger_speed_percent: int, in_range: float = 0.0
  ) -> None:
    """Releases the plate after a GraspPlate command.

    Low level method. Use `place_plate` instead for typical pick-and-place operations.

    Opens (or closes) the gripper to the specified width and cancels the force limit
    once the plate is released to avoid applying an excessive force to the plate.

    Args:
      open_width_mm: Open width in mm.
      finger_speed_percent: Percent speed to open fingers. 1 to 100.
      in_range: Optional. The standard InRange profile property for the gripper open move.
        If omitted, a zero value is assumed.

    Raises:
      ValueError: If finger_speed_percent is not between 1 and 100.
    """
    if not (1 <= finger_speed_percent <= 100):
      raise ValueError("Finger speed percent must be between 1 and 100")

    await self.send_command(f"ReleasePlate {open_width_mm} {finger_speed_percent} {in_range}")

  async def is_gripper_closed(self) -> bool:
    """(Single Gripper Only) Tests if the gripper is fully closed by checking the end-of-travel sensor.

    Returns:
      For standard gripper: True if the gripper is within 2mm of fully closed, otherwise False.
    """
    if self._is_dual_gripper:
      raise ValueError("IsGripperClosed command is only valid for single gripper robots.")
    response = await self.send_command("IsFullyClosed")
    return int(response) == -1

  async def are_grippers_closed(self) -> tuple[bool, bool]:
    """(Dual Gripper Only) Tests if each gripper is fully closed by checking the end-of-travel sensors."""
    if not self._is_dual_gripper:
      raise ValueError("AreGrippersClosed command is only valid for dual gripper robots.")
    response = await self.send_command("IsFullyClosed")
    ret_int = int(response)
    gripper_1_closed = (ret_int & 1) != 0
    gripper_2_closed = (ret_int & 2) != 0
    return (gripper_1_closed, gripper_2_closed)

  async def set_active_gripper(
    self, gripper_id: int, spin_mode: int = 0, profile_index: Optional[int] = None
  ) -> None:
    """(Dual Gripper Only) Sets the currently active gripper and modifies the tool reference frame.

    Args:
      gripper_id: Gripper ID, either 1 or 2. Determines which gripper is set to active.
      spin_mode: Optional spin mode.
        0 or omitted = do not rotate the gripper 180deg immediately.
        1 = Rotate gripper 180deg immediately.
      profile_index: Profile Index to use for spin motion.

    Raises:
      ValueError: If gripper_id is not 1 or 2, or if spin_mode is not 0 or 1.
    """
    if gripper_id not in [1, 2]:
      raise ValueError("Gripper ID must be 1 or 2")

    if spin_mode not in [0, 1]:
      raise ValueError("Spin mode must be 0 or 1")

    if profile_index is not None:
      await self.send_command(f"SetActiveGripper {gripper_id} {spin_mode} {profile_index}")
    else:
      await self.send_command(f"SetActiveGripper {gripper_id} {spin_mode}")

  async def get_active_gripper(self) -> int:
    """(Dual Gripper Only) Returns the currently active gripper.

    Returns:
      1 if Gripper A is active, 2 if Gripper B is active.
    """
    if not self._is_dual_gripper:
      raise ValueError("GetActiveGripper command is only valid for dual gripper robots.")
    response = await self.send_command("GetActiveGripper")
    return int(response)

  async def freedrive_mode(self, free_axes: List[int]) -> None:
    """Enter freedrive mode, allowing manual movement of the specified joints.

    The robot must be attached to enter free mode.

    Args:
      free_axes: List of joint indices to free. Use [0] for all axes.
    """
    for axis in free_axes:
      await self.send_command(f"freemode {axis}")

  async def end_freedrive_mode(self) -> None:
    """Exit freedrive mode for all axes."""
    await self.send_command("freemode -1")

  async def pick_plate_from_stored_position(
    self,
    position_id: int,
    horizontal_compliance: bool = False,
    horizontal_compliance_torque: int = 0,
  ):
    """Pick an item at the specified position ID.

    Args:
      position_id: The ID of the position where the plate should be picked.
      horizontal_compliance: enable horizontal compliance while closing the gripper to allow centering around the plate.
      horizontal_compliance_torque: The % of the original horizontal holding torque to be retained during compliance. If omitted, 0 is used.
    """
    horizontal_compliance_int = 1 if horizontal_compliance else 0
    ret_code = await self.send_command(
      f"pickplate {position_id} {horizontal_compliance_int} {horizontal_compliance_torque}"
    )
    if ret_code == "0":
      raise PreciseFlexError(-1, "the force-controlled gripper detected no plate present.")

  async def place_plate_to_stored_position(
    self,
    position_id: int,
    horizontal_compliance: bool = False,
    horizontal_compliance_torque: int = 0,
  ):
    """Place an item at the specified position ID.

    Args:
      position_id: The ID of the position where the plate should be placed.
      horizontal_compliance: enable horizontal compliance during the move to place the plate, to allow centering in the fixture.
      horizontal_compliance_torque: The % of the original horizontal holding torque to be retained during compliance.  If omitted, 0 is used.
    """
    horizontal_compliance_int = 1 if horizontal_compliance else 0
    await self.send_command(
      f"placeplate {position_id} {horizontal_compliance_int} {horizontal_compliance_torque}"
    )

  async def teach_position(self, position_id: int, z_clearance: float = 50.0):
    """Sets the plate location to the current robot position and configuration.  The location is saved as Cartesian coordinates.

    Args:
      position_id: The ID of the position to be taught.
      z_clearance: Optional.  The Z Clearance value. If omitted, a value of 50 is used.  Z clearance must be high enough to withdraw the gripper.
    """
    await self.send_command(f"teachplate {position_id} {z_clearance}")

  def _parse_xyz_response(
    self, parts: List[str]
  ) -> tuple[float, float, float, float, float, float]:
    if len(parts) != 6:
      raise PreciseFlexError(-1, "Unexpected response format for Cartesian coordinates.")

    x = float(parts[0])
    y = float(parts[1])
    z = float(parts[2])
    yaw = float(parts[3])
    pitch = float(parts[4])
    roll = float(parts[5])

    return (x, y, z, yaw, pitch, roll)

  def _parse_angles_response(self, parts: List[str]) -> Dict[int, float]:
    """Parse angle values from a response string.

    For self._has_rail=True:  wire order is [base, shoulder, elbow, wrist, gripper, rail]
    For self._has_rail=False: wire order is [base, shoulder, elbow, wrist, gripper]
    """

    if len(parts) < 3:
      raise PreciseFlexError(-1, "Unexpected response format for angles.")

    if self._has_rail:
      return {
        PFAxis.RAIL: float(parts[5]) if len(parts) > 5 else 0.0,
        PFAxis.BASE: float(parts[0]),
        PFAxis.SHOULDER: float(parts[1]),
        PFAxis.ELBOW: float(parts[2]),
        PFAxis.WRIST: float(parts[3]) if len(parts) > 3 else 0.0,
        PFAxis.GRIPPER: float(parts[4]) if len(parts) > 4 else 0.0,
      }

    return {
      PFAxis.RAIL: 0.0,
      PFAxis.BASE: float(parts[0]),
      PFAxis.SHOULDER: float(parts[1]),
      PFAxis.ELBOW: float(parts[2]) if len(parts) > 2 else 0.0,
      PFAxis.WRIST: float(parts[3]) if len(parts) > 3 else 0.0,
      PFAxis.GRIPPER: float(parts[4]) if len(parts) > 4 else 0.0,
    }



import contextlib
import sys
from typing import TYPE_CHECKING, Callable, Optional, cast

from pylabrobot.resources.errors import HasTipError, NoTipError
from pylabrobot.resources.tip import Tip
from pylabrobot.serializer import SerializableMixin, deserialize

if TYPE_CHECKING:
  from pylabrobot.resources.tip_rack import TipSpot


this = sys.modules[__name__]
this.tip_tracking_enabled = False  # type: ignore


def set_tip_tracking(enabled: bool):
  this.tip_tracking_enabled = enabled  # type: ignore


def does_tip_tracking() -> bool:
  return this.tip_tracking_enabled  # type: ignore


@contextlib.contextmanager
def no_tip_tracking():
  old_value = this.tip_tracking_enabled
  this.tip_tracking_enabled = False  # type: ignore
  yield
  this.tip_tracking_enabled = old_value  # type: ignore


TrackerCallback = Callable[[], None]


class TipTracker(SerializableMixin):
  """A tip tracker tracks tip operations and raises errors if the tip operations are invalid."""

  def __init__(self, thing: str):
    self.thing = thing
    self._is_disabled = False
    self._tip: Optional["Tip"] = None
    self._pending_tip: Optional["Tip"] = None
    self._tip_origin: Optional["TipSpot"] = None  # not currently in a transaction, do we need that?

    self._callback: Optional[TrackerCallback] = None

  @property
  def is_disabled(self) -> bool:
    return self._is_disabled

  @property
  def has_tip(self) -> bool:
    """Whether the tip tracker has a tip. Note that this includes pending operations."""
    return self._pending_tip is not None

  def get_tip(self) -> "Tip":
    """Get the tip. Note that does includes pending operations.

    Raises:
      NoTipError: If the tip spot does not have a tip.
    """

    if self._tip is None:
      raise NoTipError(f"{self.thing} does not have a tip.")
    return self._tip

  def disable(self) -> None:
    """Disable the tip tracker."""
    self._is_disabled = True

  def enable(self) -> None:
    """Enable the tip tracker."""
    self._is_disabled = False

  def add_tip(
    self,
    tip: Tip,
    origin: Optional["TipSpot"] = None,
    commit: bool = True,
  ) -> None:
    """Update the pending state with the operation, if the operation is valid.

    Args:
      tip: The tip to add.
      commit: Whether to commit the operation immediately. If `False`, the operation will be
        committed later with `commit()` or rolled back with `rollback()`.
    """
    if self.is_disabled:
      raise RuntimeError("Tip tracker is disabled. Call `enable()`.")
    if self._pending_tip is not None:
      raise HasTipError(f"{self.thing} already has a tip.")
    self._pending_tip = tip

    self._tip_origin = origin

    if commit:
      self.commit()

  def remove_tip(self, commit: bool = False) -> None:
    """Update the pending state with the operation, if the operation is valid"""
    if self.is_disabled:
      raise RuntimeError("Tip tracker is disabled. Call `enable()`.")
    if self._pending_tip is None:
      raise NoTipError(f"{self.thing} does not have a tip.")
    self._pending_tip = None

    if commit:
      self.commit()

  def commit(self) -> None:
    """Commit the pending operations."""
    self._tip = self._pending_tip
    # Propagate state-update callback to the tip's volume tracker
    if self._tip is not None and self._callback is not None:
      self._tip.tracker.register_callback(self._callback)
    if self._callback is not None:
      self._callback()

  def rollback(self) -> None:
    """Rollback the pending operations."""
    assert not self.is_disabled, "Tip tracker is disabled. Call `enable()`."
    self._pending_tip = self._tip

  def clear(self) -> None:
    """Clear the history."""
    self._tip = None
    self._pending_tip = None

  def serialize(self) -> dict:
    """Serialize the state of the tip tracker."""
    return {
      "tip": self._tip.serialize() if self._tip is not None else None,
      "tip_state": self._tip.tracker.serialize() if self._tip is not None else None,
      "pending_tip": self._pending_tip.serialize() if self._pending_tip is not None else None,
    }

  def load_state(self, state: dict) -> None:
    """Load a saved tip tracker state."""

    self._tip = cast(Optional[Tip], deserialize(state.get("tip")))
    self._pending_tip = cast(Optional[Tip], deserialize(state.get("pending_tip")))

  def get_tip_origin(self) -> Optional["TipSpot"]:
    """Get the origin of the current tip, if known."""
    return self._tip_origin

  def __repr__(self) -> str:
    return (
      f"TipTracker({self.thing}, is_disabled={self.is_disabled}, has_tip={self.has_tip}"
      + f" tip={self._tip}, pending_tip={self._pending_tip})"
    )

  def register_callback(self, callback: TrackerCallback) -> None:
    self._callback = callback



import json
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union

from pylabrobot.__version__ import __version__
from pylabrobot.io.errors import ValidationError

_capture_or_validation_active = False


def get_capture_or_validation_active() -> bool:
  return _capture_or_validation_active


@dataclass
class Command:
  module: str
  device_id: str
  action: str


class _CaptureWriter:
  def __init__(self):
    self._path = None
    self._tempfile = None

  def start(self, path: Path):
    if self._tempfile is not None:
      raise RuntimeError("io capture already active")
    self._path = path

    self._tempfile = tempfile.NamedTemporaryFile(delete=False)
    self._tempfile.write(b'{\n  "version": "')
    self._tempfile.write(__version__.encode("utf-8"))
    self._tempfile.write(b'",\n')
    self._tempfile.write(b'  "commands": [\n')
    self._tempfile.flush()

    global _capture_or_validation_active
    _capture_or_validation_active = True

  def record(self, command: Command):
    if self._tempfile is not None:
      encoded_command = json.dumps(command.__dict__, indent=2).encode()
      # add 4 spaces to each line
      encoded_command = b"    " + encoded_command.replace(b"\n", b"\n    ")
      self._tempfile.write(encoded_command)
      self._tempfile.write(b",\n")
      self._tempfile.flush()

  def stop(self):
    if self._path is None or self._tempfile is None:
      raise RuntimeError("io capture not active. Call start() first.")

    self._tempfile.seek(self._tempfile.tell() - 2)
    # if previous line ends with a comma, delete it
    if self._tempfile.read(1) == b",":
      self._tempfile.seek(self._tempfile.tell() - 1)
      self._tempfile.write(b"\n")
      self._tempfile.write(b"  ]\n}")
    else:
      self._tempfile.write(b"]\n}")
    self._tempfile.flush()
    self._tempfile.seek(0)

    with open(self._path, "wb") as f:
      f.write(self._tempfile.read())

    print(f"Validation file written to {self._path}")

    self._path = None

    global _capture_or_validation_active
    _capture_or_validation_active = False

    self._tempfile = None

  @property
  def capture_active(self):
    return self._tempfile is not None


class CaptureReader:
  def __init__(self, path: str):
    self.path = path
    self.commands: List[dict] = []
    with open(path, "r") as f:
      data = json.load(f)
      for c in data["commands"]:
        self.commands.append(c)
    self._command_idx = 0

  def start(self):
    global _capture_or_validation_active
    _capture_or_validation_active = True

  def next_command(self) -> dict:
    command = self.commands[self._command_idx]
    self._command_idx += 1
    return command

  def done(self):
    if self._command_idx < len(self.commands):
      left = len(self.commands) - self._command_idx
      next_command = self.commands[self._command_idx]
      raise ValidationError(
        f"Log file not fully read, {left} lines left. First command: {next_command}"
      )
    print("Validation successful!")
    self.reset()

  def reset(self):
    self._command_idx = 0

    global _capture_or_validation_active
    _capture_or_validation_active = True


capturer = _CaptureWriter()


def start_capture(fp: Union[Path, str] = Path("./validation")):
  """Start capturing all IO events to log file."""
  if not isinstance(fp, Path):
    fp = Path(fp)
  if fp.is_dir():
    raise ValueError("Path is a directory, please provide a file path.")
  capturer.start(fp)


def stop_capture():
  """Stop capturing all IO events to log file."""
  capturer.stop()



from typing import Dict, Tuple, Type

from pylabrobot.storage.liconic.constants import ControllerError, HandlingError


class LiconicControllerRelayError(Exception):
  pass


class LiconicControllerCommandError(Exception):
  pass


class LiconicControllerProgramError(Exception):
  pass


class LiconicControllerHardwareError(Exception):
  pass


class LiconicControllerWriteProtectedError(Exception):
  pass


class LiconicControllerBaseUnitError(Exception):
  pass


controller_error_map: Dict[ControllerError, Tuple[Type[Exception], str]] = {
  ControllerError.RELAY_ERROR: (
    LiconicControllerRelayError,
    "Controller system error. Undefined timer, counter, data memory, check if requested unit is valid",
  ),
  ControllerError.COMMAND_ERROR: (
    LiconicControllerCommandError,
    "Controller system error. Invalid command, check if communication is opened by CR, check command sent to controller, check for interruptions during string transmission",
  ),
  ControllerError.PROGRAM_ERROR: (
    LiconicControllerProgramError,
    "Controller system error. Firmware lost, reprogram controller",
  ),
  ControllerError.HARDWARE_ERROR: (
    LiconicControllerHardwareError,
    "Controller hardware error, turn controller ON/OFF, controller is faulty has to be replaced",
  ),
  ControllerError.WRITE_PROTECTED_ERROR: (
    LiconicControllerWriteProtectedError,
    "Controller system error. Unauthorized Access",
  ),
  ControllerError.BASE_UNIT_ERROR: (
    LiconicControllerBaseUnitError,
    "Controller system error. Unauthorized Access",
  ),
}


class LiconicHandlerPlateRemoveError(Exception):
  pass


class LiconicHandlerBarcodeReadError(Exception):
  pass


class LiconicHandlerPlatePlaceError(Exception):
  pass


class LiconicHandlerPlateSetError(Exception):
  pass


class LiconicHandlerPlateGetError(Exception):
  pass


class LiconicHandlerImportPlateError(Exception):
  pass


class LiconicHandlerExportPlateError(Exception):
  pass


class LiconicHandlerGeneralError(Exception):
  pass


handler_error_map: Dict[HandlingError, Tuple[Type[Exception], str]] = {
  HandlingError.GENERAL_HANDLING_ERROR: (
    LiconicHandlerGeneralError,
    "Handling action could not be performed in time",
  ),
  HandlingError.GATE_OPEN_ERROR: (
    LiconicHandlerGeneralError,
    "Gate could not reach upper position or Gate did not reach upper position in time",
  ),
  HandlingError.GATE_CLOSE_ERROR: (
    LiconicHandlerGeneralError,
    "Gate could not reach lower position or Gate did not reach lower position in time",
  ),
  HandlingError.GENERAL_LIFT_POSITIONING_ERROR: (
    LiconicHandlerGeneralError,
    "Handler-Lift could not reach desired level position or does not move",
  ),
  HandlingError.USER_ACCESS_ERROR: (
    LiconicHandlerGeneralError,
    "Unauthorized user access in combination with manual rotation of carrousel",
  ),
  HandlingError.STACKER_SLOT_ERROR: (
    LiconicHandlerGeneralError,
    "Stacker slot cannot be reached",
  ),
  HandlingError.REMOTE_ACCESS_LEVEL_ERROR: (
    LiconicHandlerGeneralError,
    "Undefined stacker level has been requested",
  ),
  HandlingError.PLATE_TRANSFER_DETECTION_ERROR: (
    LiconicHandlerGeneralError,
    "Export operation while plate is on transfer station",
  ),
  HandlingError.LIFT_INITIALIZATION_ERROR: (
    LiconicHandlerGeneralError,
    "Lift could not be initialized",
  ),
  HandlingError.PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerGeneralError,
    "Trying to load a plate, when a plate is already on the shovel",
  ),
  HandlingError.NO_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerGeneralError,
    "Trying to remove or place plate with no plate on the shovel",
  ),
  HandlingError.NO_RECOVERY: (
    LiconicHandlerGeneralError,
    "Recovery was not possible",
  ),
  HandlingError.IMPORT_PLATE_STACKER_POSITIONING_ERROR: (
    LiconicHandlerImportPlateError,
    "Carrousel could not reach desired radial position during Import Plate procedure or Lift could not reach transfer level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_HANDLER_TRANSFER_TURN_OUT_ERROR: (
    LiconicHandlerImportPlateError,
    "Handler could not reach outer turn position at transfer level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_SHOVEL_TRANSFER_OUTER_ERROR: (
    LiconicHandlerImportPlateError,
    "Shovel could not reach outer position at transfer level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_LIFT_TRANSFER_ERROR: (
    LiconicHandlerImportPlateError,
    "Lift did not reach upper pick position at transfer level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_SHOVEL_TRANSFER_INNER_ERROR: (
    LiconicHandlerImportPlateError,
    "Shovel could not reach inner position at transfer level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_HANDLER_TRANSFER_TURN_IN_ERROR: (
    LiconicHandlerImportPlateError,
    "Handler could not reach inner turn position at transfer level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_LIFT_STACKER_TRAVEL_ERROR: (
    LiconicHandlerImportPlateError,
    "Lift could not reach desired stacker level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_SHOVEL_STACKER_FRONT_ERROR: (
    LiconicHandlerImportPlateError,
    "Shovel could not reach front position on stacker access during Plate Import procedure.",
  ),
  HandlingError.IMPORT_PLATE_LIFT_STACKER_PLACE_ERROR: (
    LiconicHandlerImportPlateError,
    "Lift could not reach stacker place level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_SHOVEL_STACKER_INNER_ERROR: (
    LiconicHandlerImportPlateError,
    "Shovel could not reach inner position at stacker plate placement during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_LIFT_TRAVEL_BACK_ERROR: (
    LiconicHandlerImportPlateError,
    "Lift could not reach zero level during Import Plate procedure.",
  ),
  HandlingError.IMPORT_PLATE_LIFT_INIT_ERROR: (
    LiconicHandlerImportPlateError,
    "Lift could not be initialized after Import Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_LIFT_STACKER_TRAVEL_ERROR: (
    LiconicHandlerExportPlateError,
    "Carrousel could not reach desired radial position during Export Plate procedure or Lift could not reach desired stacker level during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_SHOVEL_STACKER_FRONT_ERROR: (
    LiconicHandlerExportPlateError,
    "Shovel could not reach front position on stacker access during Plate Export procedure.",
  ),
  HandlingError.EXPORT_PLATE_LIFT_STACKER_IMPORT_ERROR: (
    LiconicHandlerExportPlateError,
    "Lift could not reach stacker pick level during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_SHOVEL_STACKER_INNER_ERROR: (
    LiconicHandlerExportPlateError,
    "Shovel could not reach inner position at stacker plate pick during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_LIFT_TRANSFER_POSITIONING_ERROR: (
    LiconicHandlerExportPlateError,
    "Lift could not reach transfer level during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_HANDLER_TRANSFER_TURN_OUT_ERROR: (
    LiconicHandlerExportPlateError,
    "Handler could not reach outer turn position at transfer level during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_SHOVEL_TRANSFER_OUTER_ERROR: (
    LiconicHandlerExportPlateError,
    "Shovel could not reach outer position at transfer level during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_LIFT_TRANSFER_PLACE_ERROR: (
    LiconicHandlerExportPlateError,
    "Lift did not reach lower place position at transfer level during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_SHOVEL_TRANSFER_INNER_ERROR: (
    LiconicHandlerExportPlateError,
    "Shovel could not reach inner position at transfer level during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_HANDLER_TRANSFER_TURN_IN_ERROR: (
    LiconicHandlerExportPlateError,
    "Handler could not reach inner turn position at transfer level during Export Plate procedure",
  ),
  HandlingError.EXPORT_PLATE_LIFT_TRAVEL_BACK_ERROR: (
    LiconicHandlerExportPlateError,
    "Lift could not reach Zero position during Export Plate procedure.",
  ),
  HandlingError.EXPORT_PLATE_LIFT_INITIALIZING_ERROR: (
    LiconicHandlerExportPlateError,
    "Lift could not be initialized after Export Plate procedure.",
  ),
  HandlingError.PLATE_REMOVE_GENERAL_HANDLING_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Handling action could not be performed in time.",
  ),
  HandlingError.PLATE_REMOVE_GATE_OPEN_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Gate could not reach upper position or Gate did not reach upper position in time",
  ),
  HandlingError.PLATE_REMOVE_GATE_CLOSE_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Gate could not reach lower position or Gate did not reach lower position in time",
  ),
  HandlingError.PLATE_REMOVE_GENERAL_LIFT_POSITIONING_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Handler-Lift could not reach desired level position or does not move",
  ),
  HandlingError.PLATE_REMOVE_USER_ACCESS_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Unauthorized user access in combination with manual rotation of carrousel",
  ),
  HandlingError.PLATE_REMOVE_STACKER_SLOT_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Stacker slot cannot be reached",
  ),
  HandlingError.PLATE_REMOVE_REMOTE_ACCESS_LEVEL_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Undefined stacker level has been requested",
  ),
  HandlingError.PLATE_REMOVE_PLATE_TRANSFER_DETECTION_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Export operation while plate is on transfer station",
  ),
  HandlingError.PLATE_REMOVE_LIFT_INITIALIZATION_ERROR: (
    LiconicHandlerPlateRemoveError,
    "Lift could not be initialized",
  ),
  HandlingError.PLATE_REMOVE_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerPlateRemoveError,
    "Trying to load a plate, when a plate is already on the shovel",
  ),
  HandlingError.PLATE_REMOVE_NO_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerPlateRemoveError,
    "Trying to remove or place plate with no plate on the shovel",
  ),
  HandlingError.PLATE_REMOVE_NO_RECOVERY: (
    LiconicHandlerPlateRemoveError,
    "Recovery was not possible",
  ),
  HandlingError.BARCODE_READ_GENERAL_HANDLING_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Handling action could not be performed in time.",
  ),
  HandlingError.BARCODE_READ_GATE_OPEN_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Gate could not reach upper position or Gate did not reach upper position in time",
  ),
  HandlingError.BARCODE_READ_GATE_CLOSE_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Gate could not reach lower position or Gate did not reach lower position in time",
  ),
  HandlingError.BARCODE_READ_GENERAL_LIFT_POSITIONING_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Handler-Lift could not reach desired level position or does not move",
  ),
  HandlingError.BARCODE_READ_USER_ACCESS_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Unauthorized user access in combination with manual rotation of carrousel",
  ),
  HandlingError.BARCODE_READ_STACKER_SLOT_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Stacker slot cannot be reached",
  ),
  HandlingError.BARCODE_READ_REMOTE_ACCESS_LEVEL_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Undefined stacker level has been requested",
  ),
  HandlingError.BARCODE_READ_PLATE_TRANSFER_DETECTION_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Export operation while plate is on transfer station",
  ),
  HandlingError.BARCODE_READ_LIFT_INITIALIZATION_ERROR: (
    LiconicHandlerBarcodeReadError,
    "Lift could not be initialized",
  ),
  HandlingError.BARCODE_READ_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerBarcodeReadError,
    "Trying to load a plate, when a plate is already on the shovel",
  ),
  HandlingError.BARCODE_READ_NO_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerBarcodeReadError,
    "Trying to remove or place plate with no plate on the shovel",
  ),
  HandlingError.BARCODE_READ_NO_RECOVERY: (
    LiconicHandlerBarcodeReadError,
    "Recovery was not possible",
  ),
  HandlingError.PLATE_PLACE_GENERAL_HANDLING_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Handling action could not be performed in time.",
  ),
  HandlingError.PLATE_PLACE_GATE_OPEN_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Gate could not reach upper position or Gate did not reach upper position in time",
  ),
  HandlingError.PLATE_PLACE_GATE_CLOSE_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Gate could not reach lower position or Gate did not reach lower position in time",
  ),
  HandlingError.PLATE_PLACE_GENERAL_LIFT_POSITIONING_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Handler-Lift could not reach desired level position or does not move",
  ),
  HandlingError.PLATE_PLACE_USER_ACCESS_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Unauthorized user access in combination with manual rotation of carrousel",
  ),
  HandlingError.PLATE_PLACE_STACKER_SLOT_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Stacker slot cannot be reached",
  ),
  HandlingError.PLATE_PLACE_REMOTE_ACCESS_LEVEL_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Undefined stacker level has been requested",
  ),
  HandlingError.PLATE_PLACE_PLATE_TRANSFER_DETECTION_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Export operation while plate is on transfer station",
  ),
  HandlingError.PLATE_PLACE_LIFT_INITIALIZATION_ERROR: (
    LiconicHandlerPlatePlaceError,
    "Lift could not be initialized",
  ),
  HandlingError.PLATE_PLACE_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerPlatePlaceError,
    "Trying to load a plate, when a plate is already on the shovel",
  ),
  HandlingError.PLATE_PLACE_NO_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerPlatePlaceError,
    "Trying to remove or place plate with no plate on the shovel",
  ),
  HandlingError.PLATE_PLACE_NO_RECOVERY: (
    LiconicHandlerPlatePlaceError,
    "Recovery was not possible",
  ),
  HandlingError.PLATE_SET_GENERAL_HANDLING_ERROR: (
    LiconicHandlerPlateSetError,
    "Handling action could not be performed in time.",
  ),
  HandlingError.PLATE_SET_GATE_OPEN_ERROR: (
    LiconicHandlerPlateSetError,
    "Gate could not reach upper position or Gate did not reach upper position in time",
  ),
  HandlingError.PLATE_SET_GATE_CLOSE_ERROR: (
    LiconicHandlerPlateSetError,
    "Gate could not reach lower position or Gate did not reach lower position in time",
  ),
  HandlingError.PLATE_SET_GENERAL_LIFT_POSITIONING_ERROR: (
    LiconicHandlerPlateSetError,
    "Handler-Lift could not reach desired level position or does not move",
  ),
  HandlingError.PLATE_SET_USER_ACCESS_ERROR: (
    LiconicHandlerPlateSetError,
    "Unauthorized user access in combination with manual rotation of carrousel",
  ),
  HandlingError.PLATE_SET_STACKER_SLOT_ERROR: (
    LiconicHandlerPlateSetError,
    "Stacker slot cannot be reached",
  ),
  HandlingError.PLATE_SET_REMOTE_ACCESS_LEVEL_ERROR: (
    LiconicHandlerPlateSetError,
    "Undefined stacker level has been requested",
  ),
  HandlingError.PLATE_SET_PLATE_TRANSFER_DETECTION_ERROR: (
    LiconicHandlerPlateSetError,
    "Export operation while plate is on transfer station",
  ),
  HandlingError.PLATE_SET_LIFT_INITIALIZATION_ERROR: (
    LiconicHandlerPlateSetError,
    "Lift could not be initialized",
  ),
  HandlingError.PLATE_SET_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerPlateSetError,
    "Trying to load a plate, when a plate is already on the shovel",
  ),
  HandlingError.PLATE_SET_NO_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerPlateSetError,
    "Trying to remove or place plate with no plate on the shovel",
  ),
  HandlingError.PLATE_SET_NO_RECOVERY: (
    LiconicHandlerPlateSetError,
    "Recovery was not possible",
  ),
  HandlingError.PLATE_GET_GENERAL_HANDLING_ERROR: (
    LiconicHandlerPlateGetError,
    "Handling action could not be performed in time.",
  ),
  HandlingError.PLATE_GET_GATE_OPEN_ERROR: (
    LiconicHandlerPlateGetError,
    "Gate could not reach upper position or Gate did not reach upper position in time",
  ),
  HandlingError.PLATE_GET_GATE_CLOSE_ERROR: (
    LiconicHandlerPlateGetError,
    "Gate could not reach lower position or Gate did not reach lower position in time",
  ),
  HandlingError.PLATE_GET_GENERAL_LIFT_POSITIONING_ERROR: (
    LiconicHandlerPlateGetError,
    "Handler-Lift could not reach desired level position or does not move",
  ),
  HandlingError.PLATE_GET_USER_ACCESS_ERROR: (
    LiconicHandlerPlateGetError,
    "Unauthorized user access in combination with manual rotation of carrousel",
  ),
  HandlingError.PLATE_GET_STACKER_SLOT_ERROR: (
    LiconicHandlerPlateGetError,
    "Stacker slot cannot be reached",
  ),
  HandlingError.PLATE_GET_REMOTE_ACCESS_LEVEL_ERROR: (
    LiconicHandlerPlateGetError,
    "Undefined stacker level has been requested",
  ),
  HandlingError.PLATE_GET_PLATE_TRANSFER_DETECTION_ERROR: (
    LiconicHandlerPlateGetError,
    "Export operation while plate is on transfer station",
  ),
  HandlingError.PLATE_GET_LIFT_INITIALIZATION_ERROR: (
    LiconicHandlerPlateGetError,
    "Lift could not be initialized",
  ),
  HandlingError.PLATE_GET_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerPlateGetError,
    "Trying to load a plate, when a plate is already on the shovel",
  ),
  HandlingError.PLATE_GET_NO_PLATE_ON_SHOVEL_DETECTION: (
    LiconicHandlerPlateGetError,
    "Trying to remove or place plate with no plate on the shovel",
  ),
  HandlingError.PLATE_GET_NO_RECOVERY: (
    LiconicHandlerPlateGetError,
    "Recovery was not possible during get plate",
  ),
}



from pylabrobot.resources.hamilton.tip_creators import HamiltonTip, TipPickupMethod, TipSize
from pylabrobot.resources.tip_rack import TipRack, TipSpot
from pylabrobot.resources.utils import create_ordered_items_2d


def imcs_tip_1000uL() -> HamiltonTip:
  """IMCS tip. Same as "Hamilton High volume (1000 µL) tip without a filter", but tips are 14mm shorter."""
  return HamiltonTip(
    has_filter=True,
    total_tip_length=95.1 - 14,  # - 14
    maximal_volume=1065,
    tip_size=TipSize.HIGH_VOLUME,
    # requires "full_blowout" pickup method from STARBackend.pick_up_tips96. Not available on pip channels.
    pickup_method=TipPickupMethod.OUT_OF_RACK,
  )


def imcs_96_tiprack_1000uL_filter(name: str, with_tips: bool = True) -> TipRack:
  """Tip racks for IMCS tips. Same as Hamilton 1000 µL filter tip racks, but tips are 14mm shorter.

  requires `"full_blowout"` pickup method from STARBackend.pick_up_tips96.
  Not available on pip channels...

  Part numbers:
  ...
  """

  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=imcs_96_tiprack_1000uL_filter.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-83.5 + 14,  # 14mm shorter
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=imcs_tip_1000uL,
    ),
    with_tips=with_tips,
  )


def imcs_tip_300uL() -> HamiltonTip:
  """IMCS tip. Same as "Hamilton Standard volume (300 µL) tip without a filter", but tips are 6mm shorter."""
  return HamiltonTip(
    has_filter=True,
    total_tip_length=59.9 - 6,  # - 6
    maximal_volume=360,
    tip_size=TipSize.STANDARD_VOLUME,
    # requires "full_blowout" pickup method from STARBackend.pick_up_tips96. Not available on pip channels.
    pickup_method=TipPickupMethod.OUT_OF_RACK,
  )


def imcs_96_tiprack_300uL_filter(name: str, with_tips: bool = True) -> TipRack:
  """Tip racks for IMCS tips. Same as Hamilton 300 µL filter tip racks, but tips are 6mm shorter.

  requires `"full_blowout"` pickup method from STARBackend.pick_up_tips96.
  Not available on pip channels...

  Part numbers:
  - 04T-H8R80A-1-5-96
  - 04T-H8R80A-1-10-96
  - 04T-H8R80P-1-5-96
  - 04T-H8R80P-1-10-96
  - 04T-H8R72-1-2-96
  - 04T-H8R72-1-5-96
  - 04T-H8R72-1-10-96
  - 04T-H8R72Q-1-10-96
  - 04T-H8R85P-1-5-96
  - 04T-H8R85P-1-10-96
  - 04T-H8R88F-1-2-96
  - 04T-H8R88F-1-5-96
  - 04T-H8R88F-1-10-96
  - 04T-H8R89-1-10-96
  - 04T-H8D20F-1A-3-96
  - 04T-H8CD20F-1A-3-96
  - 04T-H8R68-1-5-96
  - 04T-H8R73-1-10-96
  - 04T-H8R05-1-2-96
  - 04T-H8R05-1-5-96
  - 04T-H8R41-1-2-96
  - 04T-H8R53-1-2-96
  - 04T-H8R52-1-2-96
  - 04T-H8R52-1-5-96
  - 04T-H8R30-1-2-96
  - 04T-H8R30-1-5-96
  - 04T-H8R03R-1-5-96
  - 04T-H8R02R-1-2-96
  - 04T-H8R02R-1-5-96
  - 04T-I3R73-1-10-96
  """

  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=imcs_96_tiprack_300uL_filter.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-50.5 + 6,  # 6mm shorter
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=imcs_tip_300uL,
    ),
    with_tips=with_tips,
  )



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Dict, List, Optional

from pylabrobot.machines.backend import MachineBackend
from pylabrobot.plate_reading.standard import (
  Exposure,
  FocalPosition,
  Gain,
  ImagingMode,
  ImagingResult,
  Objective,
)
from pylabrobot.resources.plate import Plate
from pylabrobot.resources.well import Well


class PlateReaderBackend(MachineBackend, metaclass=ABCMeta):
  """An abstract class for a plate reader. Plate readers are devices that can read luminescence,
  absorbance, or fluorescence from a plate."""

  @abstractmethod
  async def setup(self) -> None:
    """Set up the plate reader. This should be called before any other methods."""

  @abstractmethod
  async def stop(self) -> None:
    """Close all connections to the plate reader and make sure setup() can be called again."""

  @abstractmethod
  async def open(self) -> None:
    """Open the plate reader. Also known as plate out."""

  @abstractmethod
  async def close(self, plate: Optional[Plate]) -> None:
    """Close the plate reader. Also known as plate in."""

  @abstractmethod
  async def read_luminescence(
    self, plate: Plate, wells: List[Well], focal_height: float
  ) -> List[Dict]:
    """Read the luminescence from the plate reader.

    Returns:
      A list of dictionaries, one for each measurement. Each dictionary contains:
        "time": float,
        "temperature": float,
        "data": List[List[float]]
    """

  @abstractmethod
  async def read_absorbance(self, plate: Plate, wells: List[Well], wavelength: int) -> List[Dict]:
    """Read the absorbance from the plate reader.

    Returns:
      A list of dictionaries, one for each measurement. Each dictionary contains:
        "wavelength": int,
        "time": float,
        "temperature": float,
        "data": List[List[float]]
    """

  @abstractmethod
  async def read_fluorescence(
    self,
    plate: Plate,
    wells: List[Well],
    excitation_wavelength: int,
    emission_wavelength: int,
    focal_height: float,
  ) -> List[Dict]:
    """Read the fluorescence from the plate reader.

    Returns:
      A list of dictionaries, one for each measurement. Each dictionary contains:
        "ex_wavelength": int,
        "em_wavelength": int,
        "time": float,
        "temperature": float,
        "data": List[List[float]]
    """


class ImagerBackend(MachineBackend, metaclass=ABCMeta):
  @abstractmethod
  async def capture(
    self,
    row: int,
    column: int,
    mode: ImagingMode,
    objective: Objective,
    exposure_time: Exposure,
    focal_height: FocalPosition,
    gain: Gain,
    plate: Plate,
  ) -> ImagingResult:
    """Capture an image of the plate in the specified mode."""


class ImageReaderBackend(PlateReaderBackend, ImagerBackend):
  pass



from .hamilton_decks import (
  HamiltonDeck,
  HamiltonSTARDeck,
  STARDeck,
  STARLetDeck,
)
from .mfx_carriers import *
from .mfx_modules import *
from .nimbus_decks import NimbusDeck
from .plate_adapters import *
from .plate_carriers import *
from .tip_carriers import *
from .tip_creators import *
from .tip_racks import *
from .trough_carriers import *
from .troughs import *
from .tube_carriers import *
from .vantage_decks import VantageDeck



from enum import Enum
from typing import Optional

from .base_control import BaseControl
from .spark_enums import MovementSpeed, PlatePosition


class PlateColor(Enum):
  BLACK = "BLACK"
  WHITE = "WHITE"
  TRANSPARENT = "TRANSPARENT"
  NO = "NO"


class PlateControl(BaseControl):
  """
  This class provides methods for controlling the plate transport system.
  It includes functionalities to move the plate to absolute positions or named positions,
  retrieve current positions and coordinates, and manage movement speed.
  """

  async def move_to_position(
    self, position: PlatePosition, additional_option: Optional[str] = None
  ) -> Optional[str]:
    """Moves the plate transport to a predefined position."""
    command = f"ABSOLUTE MODULE=MTP POSITION={position.value}"
    if additional_option:
      command += f" {additional_option}"
    return await self.send_command(command)

  async def move_to_coordinate(
    self, x: Optional[int] = None, y: Optional[int] = None, z: Optional[int] = None
  ) -> Optional[str]:
    """Moves the plate transport to the specified coordinates."""
    command = "ABSOLUTE MODULE=MTP"
    if x is not None:
      command += f" X={x}"
    if y is not None:
      command += f" Y={y}"
    if z is not None:
      command += f" Z={z}"
    return await self.send_command(command)

  async def get_current_position(self) -> Optional[str]:
    """Gets the current predefined position of the plate transport."""
    return await self.send_command("?ABSOLUTE MODULE=MTP POSITION")

  async def get_current_coordinates(self) -> Optional[str]:
    """Gets the current X, Y, Z coordinates of the plate transport."""
    return await self.send_command("?ABSOLUTE MODULE=MTP")

  async def get_current_coordinate(self, motor: str) -> Optional[str]:
    """Gets the current coordinate of a specific motor (X, Y, or Z)."""
    return await self.send_command(f"?ABSOLUTE MODULE=MTP {motor.upper()}")

  async def get_available_positions(self) -> Optional[str]:
    """Gets the list of available predefined positions."""
    return await self.send_command("#ABSOLUTE MODULE=MTP POSITION")

  async def get_motor_range(self, motor: str) -> Optional[str]:
    """Gets the movement range for a specific motor (X, Y, or Z)."""
    return await self.send_command(f"#ABSOLUTE MODULE=MTP {motor.upper()}")

  async def get_motor_x_range(self) -> Optional[str]:
    return await self.get_motor_range("X")

  async def get_motor_y_range(self) -> Optional[str]:
    return await self.get_motor_range("Y")

  async def get_motor_z_range(self) -> Optional[str]:
    return await self.get_motor_range("Z")

  async def get_available_speed_modes(self) -> Optional[str]:
    """Gets the available movement speed modes."""
    response = await self.send_command("#SPEED MOVEMENT MODULE=MTP")
    return response

  async def get_current_motor_speed(self) -> Optional[str]:
    """Gets the current movement speed mode."""
    return await self.send_command("?SPEED MOVEMENT MODULE=MTP")

  async def set_motor_speed(self, speed_mode: MovementSpeed) -> Optional[str]:
    """Sets the movement speed mode."""
    return await self.send_command(f"SPEED MOVEMENT={speed_mode.value} MODULE=MTP")

  async def get_stacker_sensor_column(self, column_type: str) -> Optional[str]:
    """Gets the stacker sensor column state."""
    return await self.send_command(f"?STACKER SENSOR {column_type.upper()} COLUMN")

  async def get_stacker_sensor_plate(self, column_type: str) -> Optional[str]:
    """Gets the stacker sensor plate state."""
    return await self.send_command(f"?STACKER SENSOR {column_type.upper()} PLATE")

  async def get_stacker_sensor_lift(self, column_type: str) -> Optional[str]:
    """Gets the stacker sensor lift state."""
    return await self.send_command(f"?STACKER SENSOR {column_type.upper()} LIFT")

  async def stacker_stack_plate(
    self, column_type: str, plate_height: int, skirt_height: int
  ) -> Optional[str]:
    """Commands the stacker to stack a plate."""
    return await self.send_command(
      f"STACKER STACK COLUMN={column_type.upper()} PLATEHEIGHT={plate_height} "
      f"SKIRTHEIGHT={skirt_height}"
    )

  async def stacker_get_plate(
    self,
    column_type: str,
    plate_height: int,
    skirt_height: int,
    color: PlateColor = PlateColor.NO,
  ) -> Optional[str]:
    """Commands the stacker to get a plate."""
    return await self.send_command(
      f"STACKER GET COLUMN={column_type.upper()} PLATEHEIGHT={plate_height} "
      f"SKIRTHEIGHT={skirt_height} COLOUR={color.value}"
    )

  async def stacker_finish(self) -> Optional[str]:
    """Commands the stacker to finish its current operation."""
    return await self.send_command("STACKER FINISH")
